from datetime import timezone
from typing import Any, List, Literal, Optional, Union
from typing_extensions import TypeAlias
from typing import get_args, get_origin

import weaviate
from dateutil import parser
from pydantic import BaseModel, Field
from weaviate.collections import CollectionAsync
from weaviate.classes.query import Filter, Metrics, QueryReference, Sort
from weaviate.classes.aggregate import GroupByAggregate
from weaviate.collections.classes.filters import _Filters
from weaviate.collections.classes.grpc import Sorting
from weaviate.outputs.aggregate import AggregateGroupByReturn, AggregateReturn
from weaviate.outputs.query import QueryReturn
from weaviate.exceptions import (
    WeaviateBaseError,
    WeaviateQueryError,
    AuthenticationFailedError,
)

# == Define Pydantic models for structured outputs of LLMs


# -- Filters
class IntegerPropertyFilter(BaseModel):
    property_name: str
    operator: Literal["=", "!=", "<", ">", "<=", ">=", "IS_NULL"]
    value: int | bool
    length: Optional[bool] = False


class FloatPropertyFilter(BaseModel):
    property_name: str
    operator: Literal["=", "!=", "<", ">", "<=", ">=", "IS_NULL"]
    value: float | bool
    length: Optional[bool] = False


class TextPropertyFilter(BaseModel):
    property_name: str
    operator: Literal["=", "!=", "LIKE", "IS_NULL"]
    value: str | bool


class BooleanPropertyFilter(BaseModel):
    property_name: str
    operator: Literal["=", "!=", "IS_NULL"]
    value: bool


class DatePropertyFilter(BaseModel):
    property_name: str
    operator: Literal["=", "!=", "<", ">", "<=", ">=", "IS_NULL"]
    value: str | bool


class ListPropertyFilter(BaseModel):
    property_name: str
    operator: Literal["=", "!=", "CONTAINS_ANY", "CONTAINS_ALL", "IS_NULL"]
    value: list[str | int | float | bool] | bool
    length: Optional[bool] = False


class CreationTimeFilter(BaseModel):
    operator: Literal["=", "<", ">", "<=", ">="]
    value: str


# -- Aggregation Fields
class IntegerAggregation(BaseModel):
    property_name: str
    metrics: List[Literal["MIN", "MAX", "MEAN", "MEDIAN", "MODE", "SUM", "COUNT"]]


class FloatAggregation(BaseModel):
    property_name: str
    metrics: List[Literal["MIN", "MAX", "MEAN", "MEDIAN", "MODE", "SUM", "COUNT"]]


class TextAggregation(BaseModel):
    property_name: str
    metrics: List[Literal["TOP_OCCURRENCES", "COUNT"]]
    min_occurrences: Optional[int] = None


class BooleanAggregation(BaseModel):
    property_name: str
    metrics: List[
        Literal[
            "TOTAL_TRUE", "TOTAL_FALSE", "PERCENTAGE_TRUE", "PERCENTAGE_FALSE", "COUNT"
        ]
    ]


class DateAggregation(BaseModel):
    property_name: str
    metrics: List[Literal["MIN", "MAX", "MEAN", "MEDIAN", "MODE", "COUNT"]]


# -- Define FilterBucket type recursively
class FilterBucket(BaseModel):
    filters: List[
        Union[
            "FilterBucket",
            IntegerPropertyFilter,
            FloatPropertyFilter,
            TextPropertyFilter,
            BooleanPropertyFilter,
            DatePropertyFilter,
            ListPropertyFilter,
            CreationTimeFilter,
        ]
    ]
    operator: Literal["AND", "OR"]


FilterBucketType: TypeAlias = FilterBucket


# == Sorting for fetch objects
class SortBy(BaseModel):
    property_name: str
    direction: Literal["ascending", "descending"]


# == Full Query Definition
class QueryOutput(BaseModel):
    target_collections: List[str]
    search_type: Literal["hybrid", "keyword", "vector", "filter_only"]
    search_query: Optional[str] = None
    sort_by: Optional[SortBy] = None
    filter_buckets: Optional[List[FilterBucket] | FilterBucket] = None
    limit: Optional[int] = 5


class NonVectorisedQueryOutput(QueryOutput):
    search_type: Literal["keyword", "filter_only"]


# -- Full Aggregation Definition
class AggregationOutput(BaseModel):
    target_collections: List[str]
    filter_buckets: Optional[List[FilterBucket] | FilterBucket] = None
    groupby_property: Optional[str] = None
    integer_property_aggregations: Optional[List[IntegerAggregation]] = None
    float_property_aggregations: Optional[List[FloatAggregation]] = None
    text_property_aggregations: Optional[List[TextAggregation]] = None
    boolean_property_aggregations: Optional[List[BooleanAggregation]] = None
    date_property_aggregations: Optional[List[DateAggregation]] = None


class VectorisedAggregationOutput(AggregationOutput):
    search_type: Optional[Literal["vector", "hybrid"]] = Field(
        default=None,
        description=(
            "The type of search to use for the aggregation. If None, no search is used. "
        ),
    )
    search_query: Optional[str] = Field(
        default=None,
        description="The text to search for (using search). Only set this if search_type is specified.",
    )
    limit: Optional[int] = Field(
        default=100,
        description="The number of results to return. Only set this if search_type is specified.",
    )


# -- Schema for collections/data
class FieldSchema(BaseModel):
    type: Literal[
        "text", "float", "boolean", "date", "text[]", "float[]", "boolean[]", "date[]"
    ]
    groups: List[str]
    mean: float


class CollectionSchema(BaseModel):
    name: str
    length: int
    summary: str
    properties: dict[str, FieldSchema]


# -- Additional output fields
class DataDisplay(BaseModel):
    display_type: str
    summarise_items: bool


class QueryError(Exception):
    pass


def _catch_filter_errors(
    filters: list,
    collection_property_types: dict[str, str],
    collection_name: str,
    schema: dict | None = None,
):
    for filter in filters:
        if isinstance(filter, FilterBucket):
            _catch_filter_errors(
                filter.filters, collection_property_types, collection_name, schema
            )
        else:
            # check inverted index
            if schema:
                if (
                    filter.operator == "IS_NULL"
                    and not schema[collection_name]["index_properties"]["isNullIndexed"]
                ):
                    raise QueryError(
                        f"For collection '{collection_name}', attempted to filter on property "
                        f"'{filter.property_name}' using a IS_NULL filter, "
                        "but the property is not indexed for null values, so this filter is unavailable. "
                    )

            # check existence
            if filter.property_name not in collection_property_types:
                if (
                    "." in filter.property_name
                    and filter.property_name.split(".")[0] in collection_property_types
                ):
                    raise QueryError(
                        f"During the filter: {filter}"
                        f"Property '{filter.property_name}' not found in the collection '{collection_name}'. "
                        f"If you are trying to filter on a nested property of '{filter.property_name.split('.')[0]}', "
                        "you cannot do this. Object properties cannot be filtered on. "
                        "This is a Weaviate limitation."
                    )
                raise QueryError(
                    f"During the filter: {filter}"
                    f"Property '{filter.property_name}' not found in the collection '{collection_name}'. "
                    f"Available properties: {list(collection_property_types.keys())}"
                )

            # check if object
            if collection_property_types[filter.property_name].startswith("object"):
                raise QueryError(
                    f"Attempted to filter on property '{filter.property_name}', "
                    "but the property type is an object. "
                    "Object types cannot be filtered on."
                )

            if "length" in filter.model_dump() and filter.length:
                if collection_property_types[filter.property_name].endswith("[]"):
                    pass

                elif (
                    schema
                    and not schema[collection_name]["index_properties"][
                        "isLengthIndexed"
                    ]
                ):
                    raise QueryError(
                        f"For collection '{collection_name}', attempted to filter on property "
                        f"'{filter.property_name}' using a length filter, "
                        "but the property is not indexed for length values, so this filter is unavailable. "
                    )
                continue

            # get the types from the union and extract their origins for isinstance check
            union_list_types = get_args(ListPropertyFilter.__annotations__["value"])
            valid_list_types = tuple(
                get_origin(type_) or type_ for type_ in union_list_types
            )

            if not isinstance(
                filter.value, valid_list_types
            ) and collection_property_types[filter.property_name].endswith("[]"):
                if collection_property_types[filter.property_name] == "text":
                    raise QueryError(
                        f"Attempted to filter on property '{filter.property_name}' using a list filter, "
                        "but the property type is text. Text properties cannot be filtered on. "
                        "Hint: use a text filter with the 'LIKE' operator instead."
                    )
                else:
                    raise QueryError(
                        f"Attempted to filter on property '{filter.property_name}' using a list filter, "
                        "but the property type is not a list. It is a "
                        f"{collection_property_types[filter.property_name]}."
                    )

            # check number
            if (
                not isinstance(
                    filter.value, IntegerPropertyFilter.__annotations__["value"]
                )
                and collection_property_types[filter.property_name] == "int"
            ):
                if collection_property_types[filter.property_name] == "float":
                    raise QueryError(
                        f"Attempted to filter on property '{filter.property_name}' using a integer filter, "
                        f"but the property type is a {collection_property_types[filter.property_name]}. "
                        f"Filter value: {filter.value} is a {type(filter.value)}, not an integer. "
                        f"Hint: you could use {float(filter.value)} instead (may require adjusting the filter operator)."
                    )
                else:
                    raise QueryError(
                        f"Attempted to filter on property '{filter.property_name}' using a integer filter, "
                        f"but the property type is a {collection_property_types[filter.property_name]}. "
                        f"Filter value: {filter.value} is a {type(filter.value)}, not an integer. "
                    )

            # check number
            if (
                not isinstance(
                    filter.value, FloatPropertyFilter.__annotations__["value"]
                )
                and collection_property_types[filter.property_name] == "float"
            ):
                if collection_property_types[filter.property_name] == "int":
                    raise QueryError(
                        f"Attempted to filter on property '{filter.property_name}' using a float filter, "
                        f"but the property type is a {collection_property_types[filter.property_name]}. "
                        f"Filter value: {filter.value} is a {type(filter.value)}, not a float. "
                        f"Hint: you could use {int(filter.value)} instead (may require adjusting the filter operator)."
                    )
                else:
                    raise QueryError(
                        f"Attempted to filter on property '{filter.property_name}' using a float filter, "
                        f"but the property type is a {collection_property_types[filter.property_name]}. "
                        f"Filter value: {filter.value} is a {type(filter.value)}, not a float. "
                    )

            # check date
            if (
                not isinstance(
                    filter.value, DatePropertyFilter.__annotations__["value"]
                )
                and collection_property_types[filter.property_name] == "date"
            ):
                raise QueryError(
                    f"Attempted to filter on property '{filter.property_name}' using a date filter, "
                    f"but the property type is a {collection_property_types[filter.property_name]}. "
                    f"Filter value: {filter.value} is a {type(filter.value)}, not a date."
                )

            # check boolean
            if (
                not isinstance(
                    filter.value, BooleanPropertyFilter.__annotations__["value"]
                )
                and collection_property_types[filter.property_name] == "boolean"
            ):
                raise QueryError(
                    f"Attempted to filter on property '{filter.property_name}' using a boolean filter, "
                    f"but the property type is a {collection_property_types[filter.property_name]}. "
                    f"Filter value: {filter.value} is a {type(filter.value)}, not a boolean."
                )

            # check text
            if (
                not isinstance(
                    filter.value, TextPropertyFilter.__annotations__["value"]
                )
                and collection_property_types[filter.property_name] == "text"
            ):
                raise QueryError(
                    f"Attempted to filter on property '{filter.property_name}' using a text filter, "
                    f"but the property type is a {collection_property_types[filter.property_name]}. "
                    f"Filter value: {filter.value} is a {type(filter.value)}, not a text."
                )


# == Define functions for executing queries and aggregations
def _catch_typing_errors(
    tool_args: dict[str, Any],
    property_types: dict[str, dict[str, str]],
    schema: dict | None = None,
) -> None:
    """Catch typing errors and raise a QueryError."""

    # check search query
    if (
        "search_type" in tool_args
        and tool_args["search_type"] in ["hybrid", "keyword", "vector"]
        and "search_query" not in tool_args
    ):
        raise QueryError("Search query cannot be None for non-filter-only search.")

    if (
        "search_query" in tool_args
        and tool_args["search_query"] is not None
        and ("search_type" not in tool_args or tool_args["search_type"] is None)
    ):
        raise QueryError(
            "A search query was provided but no search type was provided. "
            "You can retry with a search type included. "
        )

    for collection_name, collection_property_types in property_types.items():

        aggregations = [
            arg for arg in list(tool_args.keys()) if arg.endswith("aggregations")
        ]

        if len(aggregations) > 0:
            for aggregation in aggregations:
                for agg_operator in tool_args[aggregation]:

                    if agg_operator.property_name not in collection_property_types:
                        if (
                            "." in agg_operator.property_name
                            and agg_operator.property_name.split(".")[0]
                            in collection_property_types
                        ):
                            raise QueryError(
                                f"During the aggregation '{aggregation}': "
                                f"Property '{agg_operator.property_name}' not found in the collection '{collection_name}'. "
                                f"If you are trying to aggregate on a nested property of '{agg_operator.property_name.split('.')[0]}', "
                                "you cannot do this. Object properties cannot be aggregated on. "
                                "This is a Weaviate limitation."
                            )
                        raise QueryError(
                            f"During the aggregation: {aggregation}, "
                            f"Property '{agg_operator.property_name}' not found in the collection '{collection_name}'. "
                            f"Available properties: {list(collection_property_types.keys())}"
                        )

                    # check if object
                    if collection_property_types[agg_operator.property_name].startswith(
                        "object"
                    ):
                        raise QueryError(
                            f"Attempted to filter on property '{filter.property_name}', "
                            "but the property type is an object. "
                            "Object types cannot be filtered on."
                        )

        if "filter_buckets" in tool_args:
            _catch_filter_errors(
                tool_args["filter_buckets"],
                collection_property_types,
                collection_name,
                schema,
            )


def _catch_weaviate_errors(e: WeaviateBaseError):
    if isinstance(e, WeaviateQueryError):
        if "VectorFromInput was called without vectorizer" in e.message:
            raise QueryError(
                "You are trying to do hybrid or vector search on a collection that has no vectorizer. "
                "You can only perform filter-only or keyword search on this collection. "
            )
        else:
            raise e
    elif isinstance(e, AuthenticationFailedError):
        raise QueryError(
            "Weaviate authentication failed. The user should check their API key and cluster URL or Weaviate connection details."
        )
    else:
        raise e


async def execute_weaviate_query(
    weaviate_client: weaviate.WeaviateAsyncClient,
    predicted_query: QueryOutput,
    property_types: dict[str, dict[str, str]],
    reference_property: str | None = None,
    named_vector_fields: dict[str, list[str]] | None = None,
    schema: dict | None = None,
) -> tuple[list[QueryReturn], list[str]]:
    """Execute from a QueryOutput and return response."""

    # Convert WeaviateQuery to tool args format
    tool_args: dict[str, Any] = {
        "collection_names": predicted_query.target_collections,
    }

    if predicted_query.search_query:
        tool_args["search_query"] = predicted_query.search_query

    if predicted_query.search_type:
        tool_args["search_type"] = predicted_query.search_type

    if predicted_query.filter_buckets:
        if isinstance(predicted_query.filter_buckets, list):
            tool_args["filter_buckets"] = predicted_query.filter_buckets
        elif isinstance(predicted_query.filter_buckets, FilterBucket):
            tool_args["filter_buckets"] = [predicted_query.filter_buckets]

    if predicted_query.sort_by:
        tool_args["sort_by"] = predicted_query.sort_by.model_dump()

    if predicted_query.limit:
        tool_args["limit"] = min(predicted_query.limit, 100)
    else:
        tool_args["limit"] = min(QueryOutput.model_fields["limit"].default, 100)

    if reference_property:
        tool_args["reference_property"] = reference_property

    if named_vector_fields:
        tool_args["named_vector_fields"] = {
            name: fields if fields != [] else None
            for name, fields in named_vector_fields.items()
        }

    _catch_typing_errors(tool_args, property_types, schema)

    try:
        final_responses, str_responses = await _handle_search(
            weaviate_client, tool_args
        )
    except WeaviateBaseError as e:
        _catch_weaviate_errors(e)

    return final_responses, str_responses


async def _handle_search(
    weaviate_client: weaviate.WeaviateAsyncClient, tool_args: dict
) -> tuple[list[QueryReturn], list[str]]:
    """Do vector/keyword/hybrid search from a QueryOutput."""

    collection_names = tool_args["collection_names"]
    collections = [weaviate_client.collections.get(name) for name in collection_names]

    # Build reference property
    if "reference_property" in tool_args:
        reference = QueryReference(
            link_on=tool_args["reference_property"],
            return_properties=[],
        )
    else:
        reference = None

    if "named_vector_fields" in tool_args:
        named_vector_fields = tool_args["named_vector_fields"]
    else:
        named_vector_fields = None

    # Filter and sort
    combined_filter = _build_filters(tool_args)
    sort = _build_sort(tool_args)

    # Loop over collections and execute search
    responses = []
    str_responses = []
    for i, collection in enumerate(collections):
        if tool_args["search_type"] == "filter_only":
            response = await collection.query.fetch_objects(
                limit=tool_args["limit"],
                filters=combined_filter,
                return_references=reference,
                sort=sort,
            )
        else:
            if tool_args["search_type"] == "hybrid":
                response = await collection.query.hybrid(
                    query=tool_args["search_query"],
                    limit=tool_args["limit"],
                    filters=combined_filter,
                    return_references=reference,
                    target_vector=(
                        named_vector_fields[collection.name]
                        if named_vector_fields
                        else None
                    ),
                )

            elif tool_args["search_type"] == "vector":
                response = await collection.query.near_text(
                    query=tool_args["search_query"],
                    limit=tool_args["limit"],
                    filters=combined_filter,
                    return_references=reference,
                    target_vector=(
                        named_vector_fields[collection.name]
                        if named_vector_fields
                        else None
                    ),
                )

            elif tool_args["search_type"] == "keyword":
                response = await collection.query.bm25(
                    query=tool_args["search_query"],
                    limit=tool_args["limit"],
                    filters=combined_filter,
                    return_references=reference,
                )

        str_response = _construct_string_search_query(tool_args, combined_filter)
        responses.append(response)
        str_responses.append(str_response)

    return responses, str_responses


def _build_sort(tool_args: dict) -> Sorting | None:
    """Build a sort object from a QueryOutput."""

    if "sort_by" in tool_args and tool_args["search_type"] == "filter_only":
        return Sort.by_property(
            tool_args["sort_by"]["property_name"],
            ascending=tool_args["sort_by"]["direction"] == "ascending",
        )
    return None


def _build_single_filter(
    filter: (
        IntegerPropertyFilter
        | FloatPropertyFilter
        | TextPropertyFilter
        | BooleanPropertyFilter
        | DatePropertyFilter
        | ListPropertyFilter
        | CreationTimeFilter
    ),
    reference_property: str | None = None,
) -> Filter:
    """Build a Weaviate Filter from a typed Filter object."""

    if isinstance(filter, CreationTimeFilter):
        prop_name = ""
        filter_operator = filter.operator
        filter_value = filter.value
        filter_length = False
    else:
        prop_name = filter.property_name
        filter_operator = filter.operator
        filter_value = filter.value
        if not (
            isinstance(filter, TextPropertyFilter)
            or isinstance(filter, DatePropertyFilter)
            or isinstance(filter, BooleanPropertyFilter)
        ):
            filter_length = filter.length
        else:
            filter_length = False

    if (
        isinstance(filter, FloatPropertyFilter)
        and filter_operator != "IS_NULL"
        and isinstance(filter_value, (int, float, bool))
    ):
        filter_value = float(filter_value)
    elif isinstance(filter, IntegerPropertyFilter) and isinstance(filter_value, float):
        filter_value = int(filter_value)
    elif isinstance(filter, DatePropertyFilter) and isinstance(filter_value, str):
        filter_value = parser.parse(filter_value).replace(tzinfo=timezone.utc)
    elif isinstance(filter, CreationTimeFilter) and isinstance(filter_value, str):
        filter_value = parser.parse(filter_value).replace(tzinfo=timezone.utc)

    if reference_property:
        filter_obj = Filter.by_ref(link_on=reference_property)
    else:
        filter_obj = Filter

    if isinstance(filter, CreationTimeFilter):
        filter_obj = filter_obj.by_creation_time()
    else:
        filter_obj = filter_obj.by_property(
            prop_name, length=filter_length if filter_length else False
        )

    if filter_operator == "=":
        return filter_obj.equal(filter_value)  # type: ignore
    elif filter_operator == "!=":
        return filter_obj.not_equal(filter_value)  # type: ignore
    elif filter_operator == ">":
        return filter_obj.greater_than(filter_value)  # type: ignore
    elif filter_operator == "<":
        return filter_obj.less_than(filter_value)  # type: ignore
    elif filter_operator == ">=":
        return filter_obj.greater_or_equal(filter_value)  # type: ignore
    elif filter_operator == "<=":
        return filter_obj.less_or_equal(filter_value)  # type: ignore
    elif filter_operator == "LIKE":
        return filter_obj.like(filter_value)  # type: ignore
    elif filter_operator == "CONTAINS_ANY":
        return filter_obj.contains_any(filter_value)  # type: ignore
    elif filter_operator == "CONTAINS_ALL":
        return filter_obj.contains_all(filter_value)  # type: ignore
    elif filter_operator == "IS_NULL":
        return filter_obj.is_none(filter_value)  # type: ignore
    else:
        raise ValueError(f"Invalid filter operator: {filter_operator}")


def _build_filter_bucket(
    filter_bucket: FilterBucket, reference_collection: str | None = None
) -> _Filters:
    """
    Build a collection of individual filters from a FilterBucket.
    Recursively build the filters if it finds more FilterBuckets.
    """
    bucket_operator = filter_bucket.operator
    filters = []
    for filter in filter_bucket.filters:
        if isinstance(filter, FilterBucket):
            filters.append(_build_filter_bucket(filter))
        elif (
            isinstance(filter, IntegerPropertyFilter)
            or isinstance(filter, FloatPropertyFilter)
            or isinstance(filter, TextPropertyFilter)
            or isinstance(filter, BooleanPropertyFilter)
            or isinstance(filter, DatePropertyFilter)
            or isinstance(filter, ListPropertyFilter)
            or isinstance(filter, CreationTimeFilter)
        ):
            filters.append(_build_single_filter(filter, reference_collection))

    if bucket_operator == "AND":
        return Filter.all_of(filters)
    elif bucket_operator == "OR":
        return Filter.any_of(filters)
    else:
        raise ValueError(f"Invalid bucket operator: {bucket_operator}")


def _build_filters(tool_args: dict) -> _Filters | None:
    """
    Build all filters from a QueryOutput.
    """

    if "filter_buckets" not in tool_args:
        return None

    filters = []

    for bucket in tool_args["filter_buckets"]:
        bucket_filters = _build_filter_bucket(bucket)
        filters.append(bucket_filters)

    return Filter.all_of(filters)


# -- Aggregation
async def execute_weaviate_aggregation(
    weaviate_client: weaviate.WeaviateAsyncClient,
    predicted_query: AggregationOutput,
    property_types: dict[str, dict[str, str]],
    vectorised_by_collection: dict[str, bool],
    schema: dict | None = None,
) -> tuple[list[AggregateReturn], list[str]]:
    """Execute a predicted WeaviateAggregation and return formatted results."""

    # Convert WeaviateAggregation to tool args format
    tool_args: dict[str, Any] = {
        "collection_names": predicted_query.target_collections,
    }

    if isinstance(predicted_query, VectorisedAggregationOutput):
        if predicted_query.search_query:
            tool_args["search_query"] = predicted_query.search_query

        if predicted_query.search_type:
            tool_args["search_type"] = predicted_query.search_type

        if predicted_query.limit:
            tool_args["limit"] = predicted_query.limit
        else:
            tool_args["limit"] = VectorisedAggregationOutput.model_fields[
                "limit"
            ].default

    if predicted_query.filter_buckets:
        if isinstance(predicted_query.filter_buckets, list):
            tool_args["filter_buckets"] = predicted_query.filter_buckets
        elif isinstance(predicted_query.filter_buckets, FilterBucket):
            tool_args["filter_buckets"] = [predicted_query.filter_buckets]

    if predicted_query.integer_property_aggregations:
        tool_args["integer_property_aggregations"] = (
            predicted_query.integer_property_aggregations
        )

    if predicted_query.float_property_aggregations:
        tool_args["float_property_aggregations"] = (
            predicted_query.float_property_aggregations
        )

    if predicted_query.text_property_aggregations:
        tool_args["text_property_aggregations"] = (
            predicted_query.text_property_aggregations
        )

    if predicted_query.boolean_property_aggregations:
        tool_args["boolean_property_aggregations"] = (
            predicted_query.boolean_property_aggregations
        )

    if predicted_query.date_property_aggregations:
        tool_args["date_property_aggregations"] = (
            predicted_query.date_property_aggregations
        )

    if predicted_query.groupby_property:
        tool_args["groupby_property"] = predicted_query.groupby_property

    _catch_typing_errors(tool_args, property_types, schema)

    # Execute query based on type
    try:
        responses, str_responses = await _handle_aggregation_query(
            weaviate_client, tool_args, vectorised_by_collection
        )
    except WeaviateBaseError as e:
        _catch_weaviate_errors(e)

    return responses, str_responses


async def _handle_aggregation_query(
    weaviate_client: weaviate.WeaviateAsyncClient,
    tool_args: dict,
    vectorised_by_collection: dict[str, bool],
) -> tuple[list[AggregateReturn], list[str]]:
    collection_names = tool_args["collection_names"]
    collections = [weaviate_client.collections.get(name) for name in collection_names]

    agg_args = _build_aggregation_args(tool_args)
    combined_filter = _build_filters(tool_args)
    responses = []
    str_responses = []

    for collection in collections:
        if (
            "search_query" in tool_args
            and tool_args["search_query"]
            and "search_type" in tool_args
            and tool_args["search_type"]
            and vectorised_by_collection[collection.name]
        ):
            responses.append(
                await _execute_aggregation_with_search(
                    collection, tool_args, agg_args, combined_filter
                )
            )
            str_responses.append(
                _get_string_aggregation_with_search(tool_args, combined_filter)
            )
        else:
            responses.append(
                await _execute_aggregation_over_all(
                    collection, agg_args, combined_filter
                )
            )
            str_responses.append(
                _get_string_aggregation_over_all(tool_args, combined_filter)
            )

    return responses, str_responses


def _build_return_metrics(tool_args: dict) -> list[Metrics] | None:
    metrics_types = [
        "integer_property_aggregations",
        "float_property_aggregations",
        "text_property_aggregations",
        "boolean_property_aggregations",
        "date_property_aggregations",
    ]

    full_metrics = []
    for agg_type in metrics_types:
        if agg_type in tool_args:
            for agg in tool_args[agg_type]:
                prop_name = agg.property_name
                metrics = [m.upper() for m in agg.metrics]
                if prop_name:
                    if agg_type.startswith("integer"):
                        # Map to correct integer metric names
                        metric_mapping = {
                            "MEAN": "mean",
                            "SUM": "sum_",
                            "MAX": "maximum",
                            "MIN": "minimum",
                            "COUNT": "count",
                        }
                        metric_names = [
                            metric_mapping.get(m, m.lower()) for m in metrics  # type: ignore
                        ]
                        full_metrics.append(
                            Metrics(prop_name).integer(
                                **{metric_name: True for metric_name in metric_names}
                            )
                        )

                    elif agg_type.startswith("float"):
                        # Map to correct integer metric names
                        metric_mapping = {
                            "MEAN": "mean",
                            "SUM": "sum_",
                            "MAX": "maximum",
                            "MIN": "minimum",
                            "COUNT": "count",
                        }
                        metric_names = [
                            metric_mapping.get(m, m.lower()) for m in metrics  # type: ignore
                        ]
                        full_metrics.append(
                            Metrics(prop_name).number(
                                **{metric_name: True for metric_name in metric_names}
                            )
                        )

                    elif agg_type.startswith("text"):
                        # Map to correct text metric names
                        metric_mapping = {
                            "COUNT": ["count"],
                            "TOP_OCCURRENCES": [
                                "top_occurrences_count",
                                "top_occurrences_value",
                            ],
                        }
                        metric_names = [
                            metric_name
                            for m in metrics
                            for metric_name in metric_mapping.get(m, [m.lower()])
                        ]

                        if (
                            any(m.startswith("top_occurrences") for m in metric_names)
                            and agg.min_occurrences
                        ):
                            full_metrics.append(
                                Metrics(prop_name).text(
                                    limit=agg.min_occurrences,
                                    **{
                                        metric_name: True
                                        for metric_name in metric_names
                                    },
                                )
                            )
                        else:
                            full_metrics.append(
                                Metrics(prop_name).text(
                                    **{
                                        metric_name: True
                                        for metric_name in metric_names
                                    }
                                )
                            )

                    elif agg_type.startswith("boolean"):
                        metric_mapping = {
                            "PERCENTAGE_TRUE": "percentage_true",
                            "PERCENTAGE_FALSE": "percentage_false",
                            "TOTAL_TRUE": "total_true",
                            "TOTAL_FALSE": "total_false",
                            "COUNT": "count",
                        }
                        metric_names: list[str] = [
                            metric_mapping.get(m, m.lower()) for m in metrics.keys()  # type: ignore
                        ]
                        return Metrics(prop_name).boolean(
                            **{metric_name: True for metric_name in metric_names}  # type: ignore
                        )

                    elif agg_type.startswith("date"):
                        # Map to correct date metric names
                        metric_mapping = {
                            "MEAN": "mean",
                            "MAX": "maximum",
                            "MIN": "minimum",
                            "MEDIAN": "median",
                            "MODE": "mode",
                            "COUNT": "count",
                        }
                        metric_names = [
                            metric_mapping.get(m, m.lower()) for m in metrics  # type: ignore
                        ]
                        full_metrics.append(
                            Metrics(prop_name).date_(
                                **{metric_name: True for metric_name in metric_names}
                            )
                        )

    return None if full_metrics == [] else full_metrics


def _build_aggregation_args(tool_args: dict) -> dict:
    agg_args: dict = {"total_count": True}

    if "groupby_property" in tool_args:
        agg_args["group_by"] = GroupByAggregate(
            prop=tool_args["groupby_property"],
            limit=200,
        )

    agg_args["return_metrics"] = _build_return_metrics(tool_args)

    return agg_args


async def _execute_aggregation_with_search(
    collection: CollectionAsync,
    tool_args: dict,
    agg_args: dict,
    combined_filter: _Filters | None,
) -> AggregateReturn | AggregateGroupByReturn:
    if tool_args["search_type"] == "hybrid" or tool_args["search_type"] is None:
        search = collection.aggregate.hybrid
    elif tool_args["search_type"] == "vector":
        search = collection.aggregate.near_text
    else:
        raise ValueError(f"Invalid search type: {tool_args['search_type']}")

    return await search(
        query=tool_args["search_query"],
        object_limit=tool_args["limit"],
        total_count=agg_args["total_count"],
        group_by=agg_args.get("group_by"),
        return_metrics=agg_args["return_metrics"],
        filters=combined_filter,
    )


async def _execute_aggregation_over_all(
    collection: CollectionAsync,
    agg_args: dict,
    combined_filter: _Filters | None,
) -> AggregateReturn | AggregateGroupByReturn:
    return await collection.aggregate.over_all(
        total_count=agg_args["total_count"],
        group_by=agg_args.get("group_by"),
        return_metrics=agg_args.get("return_metrics"),
        filters=combined_filter,
    )


# == Strings for code display
# Repeats of the same functions above but string output
def _build_single_filter_string(
    filter: (
        IntegerPropertyFilter
        | FloatPropertyFilter
        | TextPropertyFilter
        | BooleanPropertyFilter
        | DatePropertyFilter
        | ListPropertyFilter
        | CreationTimeFilter
    ),
) -> str:

    if isinstance(filter, CreationTimeFilter):
        filter_operator = filter.operator
        filter_value = filter.value
    else:
        prop_name = filter.property_name
        filter_operator = filter.operator
        filter_value = filter.value
        if not (
            isinstance(filter, TextPropertyFilter)
            or isinstance(filter, DatePropertyFilter)
            or isinstance(filter, BooleanPropertyFilter)
        ):
            filter_length = filter.length
        else:
            filter_length = False

    # Format value based on type
    if isinstance(filter_value, str):
        formatted_value = f"'{filter_value}'"
    else:
        formatted_value = str(filter_value)

    # Map operators to their method names
    operator_map = {
        "=": "equal",
        "!=": "not_equal",
        ">": "greater_than",
        "<": "less_than",
        ">=": "greater_or_equal",
        "<=": "less_or_equal",
        "LIKE": "like",
        "CONTAINS_ANY": "contains_any",
        "CONTAINS_ALL": "contains_all",
        "IS_NULL": "is_none",
    }

    method_name = operator_map.get(filter_operator)
    if method_name is None:
        raise ValueError(f"Invalid filter operator: {filter_operator}")

    if isinstance(filter, CreationTimeFilter):
        start_string = "Filter.by_creation_time()"
    else:
        start_string = f"Filter.by_property('{prop_name}'"
        if filter_length:
            start_string += ", length=True"
        start_string += ")"

    return f"{start_string}.{method_name}({formatted_value})"


def _build_filter_bucket_string(filter_bucket: FilterBucket) -> str:
    bucket_operator = filter_bucket.operator
    filter_strings = []

    for filter in filter_bucket.filters:
        if isinstance(filter, FilterBucket):
            filter_strings.append(_build_filter_bucket_string(filter))
        elif (
            isinstance(filter, IntegerPropertyFilter)
            or isinstance(filter, FloatPropertyFilter)
            or isinstance(filter, TextPropertyFilter)
            or isinstance(filter, BooleanPropertyFilter)
            or isinstance(filter, DatePropertyFilter)
            or isinstance(filter, ListPropertyFilter)
            or isinstance(filter, CreationTimeFilter)
        ):
            filter_strings.append(_build_single_filter_string(filter))

    indented_filters = ",\n    ".join(filter_strings)

    if bucket_operator == "AND":
        return f"Filter.all_of([\n    {indented_filters}\n])"
    elif bucket_operator == "OR":
        return f"Filter.any_of([\n    {indented_filters}\n])"
    else:
        raise ValueError(f"Invalid bucket operator: {bucket_operator}")


def _build_filter_string(tool_args: dict) -> str:
    if "filter_buckets" not in tool_args:
        return "None"

    filter_strings = []

    for bucket in tool_args["filter_buckets"]:
        bucket_filter_string = _build_filter_bucket_string(bucket)
        filter_strings.append(bucket_filter_string)

    if len(filter_strings) > 1:
        indented_filters = ",\n    ".join(filter_strings)
        return f"Filter.all_of([\n    {indented_filters}\n])"
    elif len(filter_strings) == 1:
        return filter_strings[0]
    else:
        return "None"


def _construct_string_search_query(
    tool_args: dict, combined_filter: _Filters | None
) -> str:
    if tool_args["search_type"] == "hybrid":
        search_type = "hybrid"
    elif tool_args["search_type"] == "keyword":
        search_type = "bm25"
    elif tool_args["search_type"] == "vector":
        search_type = "near_text"
    elif tool_args["search_type"] == "filter_only":
        search_type = "fetch_objects"

    query_str = (
        f"query='{tool_args['search_query']}',\n    "
        if "search_query" in tool_args and tool_args["search_query"] is not None
        else ""
    )

    # Create filter string with proper indentation
    filter_string = (
        _build_filter_string(tool_args) if combined_filter is not None else ""
    )
    if filter_string and filter_string != "None":
        # If filter string contains newlines, add additional indentation
        if "\n" in filter_string:
            # Add 4 spaces to each line for proper indentation
            indented_filter = filter_string.replace("\n", "\n    ")
            filter_str = f"filters={indented_filter},\n    "
        else:
            filter_str = f"filters={filter_string},\n    "
    else:
        filter_str = ""

    limit_str = (
        f"limit={tool_args['limit']},\n    "
        if "limit" in tool_args and tool_args["limit"] is not None
        else ""
    )

    # Handle sort with proper indentation
    if "sort_by" in tool_args and tool_args["sort_by"] is not None:
        sort_str = (
            f"sort=Sort.by_property(\n"
            f"        '{tool_args['sort_by']['property_name']}',\n"
            f"        ascending={tool_args['sort_by']['direction'] == 'ascending'}\n"
            f"    )"
        )
    else:
        sort_str = ""

    # Remove trailing comma and whitespace from the last parameter
    params = f"{query_str}{filter_str}{limit_str}{sort_str}"
    if params.endswith(",\n    "):
        params = params[:-6]

    str_response = f"""collection.query.{search_type}(
    {params}
)"""
    return str_response


# -- Aggregation Strings
def _get_string_aggregation_with_search(
    tool_args: dict, combined_filter: _Filters | None
) -> str:
    metrics_str = _build_return_metrics_string(tool_args)

    if tool_args["search_type"] == "hybrid":
        search = "collection.aggregate.hybrid"
    elif tool_args["search_type"] == "vector":
        search = "collection.aggregate.near_text"

    params = []
    if "search_query" in tool_args and tool_args["search_query"]:
        params.append(f"query='{tool_args['search_query']}'")

    if "limit" in tool_args and tool_args["limit"]:
        params.append(f"object_limit={tool_args['limit']}")

    params.append("total_count=True")

    if "groupby_property" in tool_args and tool_args["groupby_property"]:
        params.append(
            f"group_by=GroupByAggregate(prop='{tool_args['groupby_property']}')"
        )

    # Create filter string with proper indentation
    filter_string = (
        _build_filter_string(tool_args) if combined_filter is not None else ""
    )
    if filter_string and filter_string != "None":
        if "\n" in filter_string:
            indented_filter = filter_string.replace("\n", "\n    ")
            params.append(f"filters={indented_filter}")
        else:
            params.append(f"filters={filter_string}")

    if metrics_str != "None":
        params.append(f"return_metrics={metrics_str}\n    ")

    # Format with proper indentation
    formatted_params = ",\n    ".join(params)

    return f"""{search}(
    {formatted_params}
)"""


def _get_string_aggregation_over_all(
    tool_args: dict, combined_filter: _Filters | None
) -> str:
    metrics_str = _build_return_metrics_string(tool_args)

    params = ["total_count=True"]

    if "groupby_property" in tool_args and tool_args["groupby_property"]:
        params.append(
            f"group_by=GroupByAggregate(prop='{tool_args['groupby_property']}')"
        )

    filter_string = (
        _build_filter_string(tool_args) if combined_filter is not None else ""
    )
    if filter_string and filter_string != "None":
        if "\n" in filter_string:
            indented_filter = filter_string.replace("\n", "\n    ")
            params.append(f"filters={indented_filter}")
        else:
            params.append(f"filters={filter_string}")

    if metrics_str != "None":
        params.append(f"return_metrics={metrics_str}")

    # Format with proper indentation
    formatted_params = ",\n    ".join(params)

    return f"""collection.aggregate.over_all(
    {formatted_params}
)"""


def _build_return_metrics_string(tool_args: dict) -> str:
    metrics_types = [
        "integer_property_aggregations",
        "float_property_aggregations",
        "text_property_aggregations",
        "boolean_property_aggregations",
        "date_property_aggregations",
    ]

    metric_strs = []
    for agg_type in metrics_types:
        if agg_type in tool_args:
            for agg in tool_args[agg_type]:
                prop_name = agg.property_name
                metrics = [m.upper() for m in agg.metrics]

                if not prop_name:
                    continue

                if agg_type.startswith("integer"):
                    # Map to correct integer metric names
                    metric_mapping = {
                        "MEAN": "mean",
                        "SUM": "sum_",
                        "MAX": "maximum",
                        "MIN": "minimum",
                        "COUNT": "count",
                        "MEDIAN": "median",
                        "MODE": "mode",
                    }
                    metric_names = [metric_mapping.get(m, m.lower()) for m in metrics]
                    metric_bools = [f"{m}=True" for m in metric_names]
                    metric_strs.append(
                        f"Metrics('{prop_name}').integer({', '.join(metric_bools)})"
                    )

                elif agg_type.startswith("float"):
                    # Map to correct float metric names
                    metric_mapping = {
                        "MEAN": "mean",
                        "SUM": "sum_",
                        "MAX": "maximum",
                        "MIN": "minimum",
                        "COUNT": "count",
                        "MEDIAN": "median",
                        "MODE": "mode",
                    }
                    metric_names = [metric_mapping.get(m, m.lower()) for m in metrics]
                    metric_bools = [f"{m}=True" for m in metric_names]
                    metric_strs.append(
                        f"Metrics('{prop_name}').number({', '.join(metric_bools)})"
                    )

                elif agg_type.startswith("text"):
                    # Map to correct text metric names
                    metric_mapping = {
                        "COUNT": ["count"],
                        "TOP_OCCURRENCES": [
                            "top_occurrences_count",
                            "top_occurrences_value",
                        ],
                    }
                    metric_names = [
                        metric_name
                        for m in metrics
                        for metric_name in metric_mapping.get(m, m.lower())  # type: ignore
                    ]
                    metric_bools = [f"{m}=True" for m in metric_names]
                    limit_part = ""
                    if (
                        "min_occurrences" in tool_args[agg_type]
                        and tool_args[agg_type].min_occurrences
                    ):
                        limit_part = f", limit={tool_args[agg_type].min_occurrences}"

                    metric_strs.append(
                        f"Metrics('{prop_name}').text({', '.join(metric_bools)}{limit_part})"
                    )

                elif agg_type.startswith("boolean"):
                    metric_mapping = {
                        "PERCENTAGE_TRUE": "percentage_true",
                        "PERCENTAGE_FALSE": "percentage_false",
                        "TOTAL_TRUE": "total_true",
                        "TOTAL_FALSE": "total_false",
                        "COUNT": "count",
                    }
                    metric_names = [metric_mapping.get(m, m.lower()) for m in metrics]
                    metric_bools = [f"{m}=True" for m in metric_names]
                    metric_strs.append(
                        f"Metrics('{prop_name}').boolean({', '.join(metric_bools)})"
                    )

                elif agg_type.startswith("date"):
                    # Map to correct date metric names
                    metric_mapping = {
                        "MEAN": "mean",
                        "MAX": "maximum",
                        "MIN": "minimum",
                        "MEDIAN": "median",
                        "MODE": "mode",
                    }
                    metric_names = [metric_mapping.get(m, m.lower()) for m in metrics]
                    metric_bools = [f"{m}=True" for m in metric_names]
                    metric_strs.append(
                        f"Metrics('{prop_name}').date({', '.join(metric_bools)})"
                    )

    if metric_strs != []:
        return "[\n        " + ",\n        ".join(metric_strs) + "\n    ]"
    else:
        return "None"



specific_return_types = {
    "conversation": (
        "Full conversations, including all messages and message authors, with timestamps and context of other messages in the conversation. "
        "This type can only be selected if there is a field that uniquely identifies what conversation each message belongs to, e.g. a 'Conversation ID', "
        "as well as a field that uniquely identifies each message within the conversation, e.g. a 'Message ID'."
    ),
    "message": (
        "Individual messages, only including the author of each individual message and timestamp, "
        "without surrounding context of other messages by different people. "
        "If the 'conversation' field is suitable, then this is also suitable by definition."
    ),
    "ticket": ("Support tickets, similar to Github issues or similar."),
    "product": (
        "Products items, so usually involving descriptions, prices, ratings, reviews, etc, but not always. "
        "Contains an image field, and space for plenty of metadata."
    ),
    "document": (
        "Text-based information, optionally with a title, author, date, and content, but not always. "
        "Ideal for any text-based information."
    ),
}

all_return_types = {
    **specific_return_types,
    "generic": (
        "Any other type of information that does not fit into the more specific categories. "
        "Contains fields for a range of different types of information, and is a good option for a wide range of data if no other display type is available. "
    ),
    "table": (
        "A table of information, with rows and columns. Used for displaying all of the data in a structured way. "
        "This is a fall-back option if not other display type is available. "
        "Alternatively, if the data or query requires a more analytical insight, this could be a good option."
    ),
}

conversation = {
    "content": "the content or text of the message, what was written. string",
    "author": "the author of the message. string",
    "timestamp": "the timestamp of the message in any format. datetime/string/other",
    "conversation_id": "the id of the conversation that the message belongs to. integer/string/other",
    "message_id": "the id of the message itself, within the conversation. integer/string/other",
}

message = {
    "content": "the content or text of the message, what was written. string",
    "author": "the author of the message. string",
    "timestamp": "the timestamp of the message in any format. datetime/string/other",
    "conversation_id": "the id of the conversation that the message belongs to. integer/string/other",
    "message_id": "the id or index of the message, used to either identify a message within a conversation or the message itself. integer/string/other",
}

ticket = {
    "title": "the title of the ticket. string",
    "subtitle": "the subtitle of the ticket. string",
    "author": "the author of the ticket. string",
    "content": "the text of the ticket. string",
    "created_at": "the timestamp of the original creation time/date of the ticket in any format. datetime/string/other",
    "updated_at": "the timestamp of the last update time/date of the ticket in any format. datetime/string/other",
    "url": "the url of the ticket. string",
    "status": "the status of the ticket. string",
    "id": "the id of the ticket. integer/string/other",
    "tags": "the tags of the ticket. list[string/other]",
    "comments": "either the comments of the ticket, or the number of comments. list[string/dict/other] / integer",
}

product = {
    "name": "the name of the product. string",
    "description": "the description of the product. string",
    "price": "the price of the product. float/integer/other",
    "category": "the category of the product. string",
    "subcategory": "the subcategory of the product. string",
    "collection": "the collection that the product belongs to. string",
    "rating": "the rating of the product. float/integer/other",
    "reviews": "the reviews of the product, or number of reviews. list[string/dict/other] / integer",
    "tags": "the tags of the product. list[string/other]",
    "url": "the url of the product. string",
    "image": "the image of the product. string/other",
    "brand": "the brand of the product. string",
    "id": "the id of the product. integer/string/other",
    "colors": "the color(s) of the product. list[string/other] / string",
    "sizes": "the size(s) of the product. list[string/other] / string",
}

document = {
    "title": "the title of the document. string",
    "author": "the author or username or creator of the document. string",
    "date": "any date or time format. datetime/string/other",
    "content": "the textual content of the document. string/other",
    "category": "some string describing the category of the document, e.g. type of something. string",
}

generic = {
    "title": "the title of the information. string",
    "subtitle": "the subtitle of the information. string",
    "content": "the content of the information. string/other",
    "url": "the url of the information. string",
    "id": "the id of the information. integer/string/other",
    "author": "the author of the information. string/other",
    "timestamp": "the timestamp of the information in any format. datetime/string/other",
    "tags": "the tags of the information. list[string/other]",
    "category": "some string describing the category of the information, e.g. type of something. string",
    "subcategory": "some string describing a nested level of category of the data. string",
}

types_dict: dict[str, dict[str, str]] = {
    "conversation": conversation,
    "message": message,
    "ticket": ticket,
    "product": product,
    "generic": generic,
    "document": document,
    "table": {},
}



from fastapi import APIRouter, Depends
from fastapi.responses import JSONResponse
from cryptography.fernet import InvalidToken

from elysia.api.api_types import InitialiseTreeData
from elysia.api.dependencies.common import get_user_manager
from elysia.api.services.user import UserManager
from elysia.api.core.log import logger
from elysia.util.client import ClientManager
from elysia.api.utils.config import Config
from elysia.api.utils.encryption import decrypt_api_keys

from weaviate.classes.query import Filter

router = APIRouter()


async def get_default_config(
    client_manager: ClientManager,
    user_id: str,
    config_collection_name: str = "ELYSIA_CONFIG__",
) -> Config | None:
    async with client_manager.connect_to_async_client() as client:
        if not await client.collections.exists(config_collection_name):
            return None

        collection = client.collections.get(config_collection_name)
        default_configs = await collection.query.fetch_objects(
            filters=Filter.all_of(
                [
                    Filter.by_property("default").equal(True),
                    Filter.by_property("user_id").equal(user_id),
                ]
            )
        )

        if len(default_configs.objects) > 0:
            default_config: dict = default_configs.objects[0].properties  # type: ignore
            try:
                default_config["settings"] = decrypt_api_keys(
                    default_config["settings"]
                )
            except InvalidToken:
                logger.warning(f"Invalid token for default config, returning None")
                return None
            return Config.from_json(default_config)
        else:
            return None


@router.post("/user/{user_id}")
async def initialise_user(
    user_id: str, user_manager: UserManager = Depends(get_user_manager)
):
    """
    Create or retrieve a user.

    Returns:
        (JSONResponse): A JSON response with the following fields:
            - error (str): Any error message (empty string if no error).
            - user_exists (bool): True if the user exists, False otherwise.
            - config (dict): The user's config.
            - frontend_config (dict): The user's frontend config.
    """
    logger.debug(f"/initialise_user API request received")

    try:

        user_exists = user_manager.user_exists(user_id)

        # if a user does not exist, create a user and set up the configs
        if not user_exists:
            await user_manager.add_user_local(
                user_id,
            )  # leave config empty to create defaults for a new user

            # find any default configs
            if user_manager.users[user_id][
                "frontend_config"
            ].save_location_client_manager.is_client:
                default_config = await get_default_config(
                    user_manager.users[user_id][
                        "frontend_config"
                    ].save_location_client_manager,
                    user_id,
                )
                if default_config:
                    await user_manager.update_config(
                        user_id,
                        config_id=default_config.id,
                        config_name=default_config.name,
                        settings=default_config.settings.to_json(),
                        style=default_config.style,
                        agent_description=default_config.agent_description,
                        end_goal=default_config.end_goal,
                        branch_initialisation=default_config.branch_initialisation,
                    )
                    logger.debug("Using default config")

        # if a user exists, get the existing configs
        user = await user_manager.get_user_local(user_id)
        config = user["tree_manager"].config.to_json()
        frontend_config = user["frontend_config"].to_json()

        correct_settings = user["tree_manager"].config.settings.check()

        logger.debug(f"--------------------------------")
        logger.debug(f"Output of init/user:")
        logger.debug(f"User ID: {user_id}")
        logger.debug(f"User exists: {user_exists}")
        logger.debug(f"Config: {config}")
        logger.debug(f"Frontend config: {frontend_config}")
        logger.debug(f"Correct settings: {correct_settings}")

    except Exception as e:
        logger.exception(f"Error in /initialise_user API")
        return JSONResponse(
            content={
                "error": str(e),
                "user_exists": None,
                "config": {},
                "frontend_config": {},
                "correct_settings": {},
            },
            status_code=500,
        )
    return JSONResponse(
        content={
            "error": "",
            "user_exists": user_exists,
            "config": config,
            "frontend_config": frontend_config,
            "correct_settings": correct_settings,
        },
        status_code=200,
    )


@router.post("/tree/{user_id}/{conversation_id}")
async def initialise_tree(
    user_id: str,
    conversation_id: str,
    data: InitialiseTreeData,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Create or retrieve a tree.

    Returns:
        (JSONResponse): A JSON response with the following fields:
            - conversation_id (str): The conversation ID.
            - tree (dict): The tree.
            - error (str): Any error message (empty string if no error).
    """
    logger.debug(f"/initialise_tree API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Conversation ID: {conversation_id}")
    logger.debug(f"Low Memory: {data.low_memory}")

    try:
        tree = await user_manager.initialise_tree(
            user_id,
            conversation_id,
            low_memory=data.low_memory,
        )
    except Exception as e:
        logger.exception(f"Error in /initialise_tree API")
        return JSONResponse(
            content={
                "conversation_id": conversation_id,
                "tree": "",
                "error": str(e),
            },
            status_code=500,
        )

    return JSONResponse(
        content={
            "conversation_id": conversation_id,
            "tree": tree.tree,
            "error": "",
        },
        status_code=200,
    )



"""
Test generic prompts and basic outputs.
Such as whether a prompt that requires searching actually does call 'query'.
Same for aggregate, text_response, summarise, itemised_summaries.
"""

import json
import pytest
from copy import deepcopy
from deepeval import evaluate, metrics
from deepeval.test_case import LLMTestCase, ToolCall

from elysia import Tree
from elysia.util.client import ClientManager
from elysia import configure

configure(logging_level="DEBUG")


def get_tools_called(tree: Tree, user_prompt: str):
    tools_called = []
    if user_prompt in tree.actions_called:
        for action in tree.actions_called[user_prompt]:
            if action["name"] in tree.tools:
                tools_called.append(
                    ToolCall(
                        name=action["name"],
                        description=tree.tools[action["name"]].description,
                        input_parameters=action["inputs"],
                        output=action["output"] if "output" in action else [],
                    )
                )
    return tools_called


async def run_tree(user_prompt: str, collection_names: list[str]) -> Tree:

    tree = Tree()

    async for result in tree.async_run(
        user_prompt,
        collection_names=collection_names,
    ):
        pass
    return tree


@pytest.mark.asyncio
async def test_query():
    """
    Test the query tool is called for a generic prompt requiring it.
    """

    # if the collection is not found, skip the test
    client_manager = ClientManager()
    client = client_manager.client
    if (
        not client.collections.exists("Example_verba_github_issues")
        or not client.collections.exists("Example_verba_email_chains")
        or not client.collections.exists("Example_verba_slack_conversations")
        or not client.collections.exists("Weather")
        or not client.collections.exists("Ml_wikipedia")
        or not client.collections.exists("Weaviate_blogs")
        or not client.collections.exists("Weaviate_documentation")
    ):
        pytest.skip("Collection not found in Weaviate")

    user_prompt = "What github issues discuss the 'PDF upload' issue?"
    tree = await run_tree(user_prompt, [])

    collection_names_to_test = [c.lower() for c in tree.tree_data.collection_names]

    # passing empty collection list should mean all collections are searched
    assert "example_verba_github_issues" in collection_names_to_test
    assert "example_verba_email_chains" in collection_names_to_test
    assert "example_verba_slack_conversations" in collection_names_to_test
    assert "weather" in collection_names_to_test
    assert "ml_wikipedia" in collection_names_to_test
    assert "weaviate_blogs" in collection_names_to_test
    assert "weaviate_documentation" in collection_names_to_test

    all_decision_history = []
    for iteration in tree.decision_history:
        all_decision_history.extend(iteration)
    assert "query" in all_decision_history

    deepeval_metric = metrics.TaskCompletionMetric(
        threshold=0.5, model="gpt-4o-mini", include_reason=True
    )

    tools_called = get_tools_called(tree, user_prompt)

    test_case = LLMTestCase(
        input=user_prompt,
        actual_output=tree.tree_data.conversation_history[-1]["content"],
        tools_called=tools_called,
    )

    res = evaluate(test_cases=[test_case], metrics=[deepeval_metric])
    for test_case in res.test_results:
        assert test_case.success, test_case.metrics_data[0].reason


@pytest.mark.asyncio
async def test_aggregate():
    """
    Test the aggregate tool is called for a generic prompt requiring it.
    """

    # if the collection is not found, skip the test
    client_manager = ClientManager()
    client = client_manager.client
    if not client.collections.exists("Ecommerce"):
        pytest.skip("Collection not found in Weaviate")

    user_prompt = "What is the average price of trousers in the ecommerce collection?"
    tree = await run_tree(
        user_prompt,
        [],
    )

    all_decision_history = []
    for iteration in tree.decision_history:
        all_decision_history.extend(iteration)
    assert "aggregate" in all_decision_history

    # Test with deepeval
    deepeval_metric = metrics.TaskCompletionMetric(
        threshold=0.5, model="gpt-4o-mini", include_reason=True
    )

    tools_called = get_tools_called(tree, user_prompt)

    test_case = LLMTestCase(
        input=user_prompt,
        actual_output=tree.tree_data.conversation_history[-1]["content"],
        tools_called=tools_called,
    )

    res = evaluate(test_cases=[test_case], metrics=[deepeval_metric])
    for test_case in res.test_results:
        assert test_case.success, test_case.metrics_data[0].reason


@pytest.mark.asyncio
async def test_text_response():
    """
    Test the text_response tool is called for a generic conversational prompt.
    """
    user_prompt = "Hey Elly what's up? What can you do?"
    tree = await run_tree(
        user_prompt,
        [],
    )

    all_decision_history = []
    for iteration in tree.decision_history:
        all_decision_history.extend(iteration)
    assert "text_response" in all_decision_history

    # Test with deepeval
    deepeval_metric = metrics.TaskCompletionMetric(
        threshold=0.5, model="gpt-4o-mini", include_reason=True
    )

    tools_called = get_tools_called(tree, user_prompt)

    test_case = LLMTestCase(
        input=user_prompt,
        actual_output=tree.tree_data.conversation_history[-1]["content"],
        tools_called=tools_called,
    )

    res = evaluate(test_cases=[test_case], metrics=[deepeval_metric])
    for test_case in res.test_results:
        assert test_case.success, test_case.metrics_data[0].reason


@pytest.mark.asyncio
async def test_summarising_output():
    """
    Test the summarise tool is called for a generic prompt requiring it.
    """

    # if the collection is not found, skip the test
    client_manager = ClientManager()
    client = client_manager.client
    if not client.collections.exists("Example_verba_github_issues"):
        pytest.skip("Collection not found in Weaviate")

    user_prompt = "Summarise the 10 most recent issues in the github issues collection"
    tree = await run_tree(
        user_prompt,
        [
            "Example_verba_github_issues",
        ],
    )

    all_decision_history = []
    for iteration in tree.decision_history:
        all_decision_history.extend(iteration)
    assert "query" in all_decision_history

    env_key = (
        "query_postprocessing"
        if "query_postprocessing" in tree.tree_data.environment.environment
        else "query"
    )

    assert (
        "Example_verba_github_issues" in tree.tree_data.environment.environment[env_key]
    )
    assert (
        "cited_summarize" in all_decision_history
        or all_decision_history[-1] == "text_response"
    )

    task_metric = metrics.TaskCompletionMetric(
        threshold=0.5, model="gpt-4o-mini", include_reason=True
    )

    tools_called = get_tools_called(tree, user_prompt)

    task_test_case = LLMTestCase(
        input=user_prompt,
        actual_output=tree.tree_data.conversation_history[-1]["content"],
        tools_called=tools_called,
    )

    res = evaluate(test_cases=[task_test_case], metrics=[task_metric])
    for test_case in res.test_results:
        assert test_case.success

    environment = [
        json.dumps(obj)
        for obj in tree.tree_data.environment.environment[env_key][
            "Example_verba_github_issues"
        ][0]["objects"]
    ]

    summarisation_metric = metrics.AnswerRelevancyMetric(
        threshold=0.5, model="gpt-4o", include_reason=True
    )
    summarisation_test_case = LLMTestCase(
        input=user_prompt,
        actual_output=tree.tree_data.conversation_history[-1]["content"],
        retrieval_context=environment,
    )

    res = evaluate(test_cases=[summarisation_test_case], metrics=[summarisation_metric])
    for test_case in res.test_results:
        assert test_case.success, test_case.metrics_data[0].reason


@pytest.mark.asyncio
async def test_itemised_summaries():
    """
    Test the itemised_summaries tool is called when specifically asked for it.
    """

    # if the collection is not found, skip the test
    client_manager = ClientManager()
    client = client_manager.client
    if not client.collections.exists("Example_verba_slack_conversations"):
        pytest.skip("Collection not found in Weaviate")

    user_prompt = (
        "Write itemised summaries for each of the 5 most recent message in slack"
    )
    tree = await run_tree(
        user_prompt,
        [
            "Example_verba_slack_conversations",
        ],
    )

    all_decision_history = []
    for iteration in tree.decision_history:
        all_decision_history.extend(iteration)
    assert "query" in all_decision_history
    assert (
        "Example_verba_slack_conversations"
        in tree.tree_data.environment.environment["query_postprocessing"]
    )
    for item in tree.tree_data.environment.environment["query_postprocessing"][
        "Example_verba_slack_conversations"
    ][0]["objects"]:
        assert "ELYSIA_SUMMARY" in item.keys() and len(item["ELYSIA_SUMMARY"]) > 0



import datetime
import os
import random
import json

from dotenv import load_dotenv
from typing import Any
from pathlib import Path
from logging import Logger

load_dotenv(override=True)

from elysia.api.services.tree import TreeManager
from elysia.objects import Update
from elysia.util.client import ClientManager
from elysia.api.core.log import logger
from elysia.api.utils.config import Config
from elysia.api.utils.config import FrontendConfig
from elysia.tree.util import get_saved_trees_weaviate


class TreeTimeoutError(Update):
    def __init__(self):
        super().__init__(
            "tree_timeout_error",
            {
                "text": "This conversation has been timed out due to inactivity. Please start a new conversation."
            },
        )


class UserTimeoutError(Update):
    def __init__(self):
        super().__init__(
            "user_timeout_error",
            {
                "text": "You have been timed out due to inactivity. Please start a new conversation."
            },
        )


async def load_frontend_config_from_file(
    user_id: str, logger: Logger
) -> FrontendConfig:
    # save frontend config locally
    try:
        elysia_package_dir = Path(__file__).parent.parent.parent  # Gets to elysia/
        config_dir = elysia_package_dir / "api" / "user_configs"
        config_file = config_dir / f"frontend_config_{user_id}.json"

        if not config_file.exists():
            return FrontendConfig(logger=logger)

        with open(config_file, "r") as f:
            fe_config = await FrontendConfig.from_json(json.load(f), logger)
            return fe_config

    except Exception as e:
        logger.error(f"Error in save_frontend_config_to_file: {str(e)}")
        return FrontendConfig(logger=logger)


class UserManager:
    """
    The UserManager is designed to manage, for each user:

    - TreeManager
    - ClientManager
    - FrontendConfig

    It contains methods for creating and updating these objects across a range of users, stored in a dictionary.
    It can be used as a dependency injection container for FastAPI.
    The user manager/tree manager is decoupled from the core of the Elysia package functionality.
    It is designed to be used to manage separate Elysia instances, set of trees (via tree managers), configs, etc.
    """

    def __init__(
        self,
        user_timeout: datetime.timedelta | int | None = None,
    ):
        """
        Args:
            user_timeout (datetime.timedelta | int | None): Optional.
                The length of time a user can be idle before being timed out.
                Defaults to 20 minutes or the value of the USER_TIMEOUT environment variable.
                If an integer is provided, it is interpreted as the number of minutes.
        """
        if user_timeout is None:
            self.user_timeout = datetime.timedelta(
                minutes=int(os.environ.get("USER_TIMEOUT", 20))
            )
        elif isinstance(user_timeout, int):
            self.user_timeout = datetime.timedelta(minutes=user_timeout)
        else:
            self.user_timeout = user_timeout

        self.manager_id = random.randint(0, 1000000)
        self.date_of_reset = None
        self.users = {}

    def user_exists(self, user_id: str):
        return user_id in self.users

    async def update_config(
        self,
        user_id: str,
        conversation_id: str | None = None,
        config_id: str | None = None,
        config_name: str | None = None,
        settings: dict[str, Any] | None = None,
        style: str | None = None,
        agent_description: str | None = None,
        end_goal: str | None = None,
        branch_initialisation: str | None = None,
    ):
        local_user = await self.get_user_local(user_id)
        local_user["tree_manager"].update_config(
            conversation_id,
            config_id,
            config_name,
            settings,
            style,
            agent_description,
            end_goal,
            branch_initialisation,
        )

        await local_user["client_manager"].reset_keys(
            wcd_url=local_user["tree_manager"].settings.WCD_URL,
            wcd_api_key=local_user["tree_manager"].settings.WCD_API_KEY,
            api_keys=local_user["tree_manager"].settings.API_KEYS,
        )

    async def update_frontend_config(
        self,
        user_id: str,
        config: dict[str, Any],
    ):
        local_user = await self.get_user_local(user_id)
        frontend_config: FrontendConfig = local_user["frontend_config"]
        await frontend_config.configure(**config)

    async def add_user_local(
        self,
        user_id: str,
        config: Config | None = None,
    ):
        """
        Add a user to the UserManager.

        Args:
            user_id (str): Required. The unique identifier for the user.
            config (Config): Required. The config for the user.
        """

        # add user if it doesn't exist
        if user_id not in self.users:
            self.users[user_id] = {}

            fe_config = await load_frontend_config_from_file(user_id, logger)

            self.users[user_id]["frontend_config"] = fe_config

            self.users[user_id]["tree_manager"] = TreeManager(
                user_id=user_id,
                config=config,
                tree_timeout=fe_config.config["tree_timeout"],
            )

            # client manager starts with env variables, when config is updated, api keys are updated
            self.users[user_id]["client_manager"] = ClientManager(
                logger=logger,
                client_timeout=fe_config.config["client_timeout"],
                settings=self.users[user_id]["tree_manager"].config.settings,
            )

    async def get_user_local(self, user_id: str):
        """
        Return a local user object.
        Will raise a ValueError if the user is not found.

        Args:
            user_id (str): Required. The unique identifier for the user.

        Returns:
            (dict): A local user object, containing a TreeManager ("tree_manager"),
                Frontend Config ("frontend_config") and ClientManager ("client_manager").
        """

        if user_id not in self.users:
            raise ValueError(
                f"User {user_id} not found. Please initialise a user first (by calling `add_user_local`)."
            )

        # update last request (adds last_request to user)
        await self.update_user_last_request(user_id)

        return self.users[user_id]

    async def get_tree(self, user_id: str, conversation_id: str):
        """
        Get a tree for a user.
        Will raise a ValueError if the user is not found, or the tree is not found.

        Args:
            user_id (str): Required. The unique identifier for the user.
            conversation_id (str): Required. The unique identifier for the conversation.

        Returns:
            (Tree): The tree.
        """
        local_user = await self.get_user_local(user_id)
        return local_user["tree_manager"].get_tree(conversation_id)

    async def check_all_trees_timeout(self):
        """
        Check all trees in all TreeManagers across all users and remove any that have not been active in the last tree_timeout.
        """
        for user_id in self.users:
            self.users[user_id]["tree_manager"].check_all_trees_timeout()

    def check_user_timeout(self, user_id: str):
        """
        Check if a user has been idle for the last user_timeout.

        Args:
            user_id (str): The user ID which contains the user.

        Returns:
            (bool): True if the user has been idle for the last user_timeout, False otherwise.
        """
        # if user not found, return True
        if user_id not in self.users:
            return True

        # Remove any trees that have not been active in the last user_timeout
        # if (
        #     "last_request" in self.users[user_id]
        #     and datetime.datetime.now() - self.users[user_id]["last_request"]
        #     > self.user_timeout
        # ):
        #     return True

        return False

    async def check_all_users_timeout(self):
        """
        Check all users in the UserManager and remove any that have not been active in the last user_timeout.
        """
        if self.user_timeout == datetime.timedelta(minutes=0):
            return

        for user_id in self.users:
            if self.check_user_timeout(user_id):
                del self.users[user_id]

    async def check_restart_clients(self):
        """
        Check all clients in all ClientManagers across all users and run the restart_client() method (for sync and async clients).
        The restart_client() methods will check if the client has been inactive for the last client_timeout minutes (set in init).
        """
        for user_id in self.users:
            if (
                "client_manager" in self.users[user_id]
                and self.users[user_id]["client_manager"].is_client
            ):
                await self.users[user_id]["client_manager"].restart_client()
                await self.users[user_id]["client_manager"].restart_async_client()

    async def close_all_clients(self):
        """
        Close all clients in all ClientManagers across all users.
        """
        for user_id in self.users:
            if "client_manager" in self.users[user_id]:
                await self.users[user_id]["client_manager"].close_clients()

    async def initialise_tree(
        self,
        user_id: str,
        conversation_id: str,
        low_memory: bool = False,
    ):
        """
        Initialises a tree for a user for an existing user at user_id.
        Requires a user to already exist in the UserManager, via `add_user_local`.
        This is a wrapper for the TreeManager.add_tree() method.

        Args:
            user_id (str): Required. The unique identifier for the user.
            conversation_id (str): Required. The unique identifier for a new conversation within the tree manager.
            low_memory (bool): Optional. Whether to use low memory mode for the tree.
                Controls the LM history being saved in the tree, and some other variables.
                Defaults to False.
        """
        # self.add_user_local(user_id)
        local_user = await self.get_user_local(user_id)
        tree_manager: TreeManager = local_user["tree_manager"]
        if not tree_manager.tree_exists(conversation_id):
            tree_manager.add_tree(
                conversation_id,
                low_memory,
            )
        return tree_manager.get_tree(conversation_id)

    async def save_tree(
        self,
        user_id: str,
        conversation_id: str,
        wcd_url: str | None = None,
        wcd_api_key: str | None = None,
    ):
        """
        Save a tree to a Weaviate instance (set in the frontend config).
        This is a wrapper for the TreeManager.save_tree_weaviate() method.

        Args:
            user_id (str): Required. The unique identifier for the user stored in the UserManager.
            conversation_id (str): Required. The unique identifier for the conversation for the user.
        """

        local_user = await self.get_user_local(user_id)
        tree_manager: TreeManager = local_user["tree_manager"]

        if wcd_url is None or wcd_api_key is None:
            save_location_client_manager = local_user[
                "frontend_config"
            ].save_location_client_manager
        else:
            save_location_client_manager = ClientManager(
                logger=logger,
                wcd_url=wcd_url,
                wcd_api_key=wcd_api_key,
            )

        await tree_manager.save_tree_weaviate(
            conversation_id, save_location_client_manager
        )

    async def check_tree_exists_weaviate(
        self,
        user_id: str,
        conversation_id: str,
        wcd_url: str | None = None,
        wcd_api_key: str | None = None,
    ):
        """
        Check if a tree exists in a Weaviate instance (set in the frontend config).

        Args:
            user_id (str): Required. The unique identifier for the user stored in the UserManager.
            conversation_id (str): Required. The unique identifier for the conversation for the user.
            wcd_url (str | None): Required. The URL of the Weaviate Cloud Database instance used to save the tree.
                Defaults to the value of the `wcd_url` setting in the frontend config.
            wcd_api_key (str | None): Required. The API key for the Weaviate Cloud Database instance used to save the tree.
                Defaults to the value of the `wcd_api_key` setting in the frontend config.

        Returns:
            (bool): True if the tree exists in the Weaviate instance, False otherwise.
        """

        local_user = await self.get_user_local(user_id)
        tree_manager: TreeManager = local_user["tree_manager"]

        if wcd_url is None or wcd_api_key is None:
            save_location_client_manager = local_user[
                "frontend_config"
            ].save_location_client_manager
        else:
            save_location_client_manager = ClientManager(
                logger=logger,
                wcd_url=wcd_url,
                wcd_api_key=wcd_api_key,
            )

        return await tree_manager.check_tree_exists_weaviate(
            conversation_id, save_location_client_manager
        )

    async def load_tree(
        self,
        user_id: str,
        conversation_id: str,
        wcd_url: str | None = None,
        wcd_api_key: str | None = None,
    ):
        """
        Load a tree from a Weaviate instance (set in the frontend config).
        This is a wrapper for the TreeManager.load_tree_weaviate() method.

        Args:
            user_id (str): Required. The unique identifier for the user stored in the UserManager.
            conversation_id (str): Required. The unique identifier for the conversation for the user.
            wcd_url (str | None): Required. The URL of the Weaviate Cloud Database instance used to save the tree.
                Defaults to the value of the `wcd_url` setting in the frontend config.
            wcd_api_key (str | None): Required. The API key for the Weaviate Cloud Database instance used to save the tree.
                Defaults to the value of the `wcd_api_key` setting in the frontend config.

        Returns:
            (list[dict]): A list of dictionaries, each containing a frontend payload that was used to generate the tree.
                The list is ordered by the time the payload was originally sent to the frontend (at the time it was saved).
        """

        local_user = await self.get_user_local(user_id)
        tree_manager: TreeManager = local_user["tree_manager"]

        if wcd_url is None or wcd_api_key is None:
            save_location_client_manager = local_user[
                "frontend_config"
            ].save_location_client_manager
        else:
            save_location_client_manager = ClientManager(
                logger=logger,
                wcd_url=wcd_url,
                wcd_api_key=wcd_api_key,
            )

        return await tree_manager.load_tree_weaviate(
            conversation_id, save_location_client_manager
        )

    async def delete_tree(
        self,
        user_id: str,
        conversation_id: str,
        wcd_url: str | None = None,
        wcd_api_key: str | None = None,
    ):
        """
        Delete a saved tree from a Weaviate instance (set in the frontend config).
        Also delete the tree from the local tree manager.
        This is a wrapper for the TreeManager.delete_tree_weaviate() method and the TreeManager.delete_tree_local() method.

        Args:
            user_id (str): Required. The unique identifier for the user stored in the UserManager.
            conversation_id (str): Required. The unique identifier for the conversation for the user.
            wcd_url (str | None): Required. The URL of the Weaviate Cloud Database instance used to save the tree.
                Defaults to the value of the `wcd_url` setting in the frontend config.
            wcd_api_key (str | None): Required. The API key for the Weaviate Cloud Database instance used to save the tree.
                Defaults to the value of the `wcd_api_key` setting in the frontend config.
        """

        local_user = await self.get_user_local(user_id)
        tree_manager: TreeManager = local_user["tree_manager"]

        if wcd_url is None or wcd_api_key is None:
            save_location_client_manager = local_user[
                "frontend_config"
            ].save_location_client_manager
        else:
            save_location_client_manager = ClientManager(
                logger=logger,
                wcd_url=wcd_url,
                wcd_api_key=wcd_api_key,
            )

        await tree_manager.delete_tree_weaviate(
            conversation_id, save_location_client_manager
        )
        tree_manager.delete_tree_local(conversation_id)

    async def get_saved_trees(
        self,
        user_id: str,
        wcd_url: str | None = None,
        wcd_api_key: str | None = None,
    ):
        """
        Get all saved trees from a Weaviate instance (set in the frontend config).

        Args:
            user_id (str): Required. The unique identifier for the user stored in the UserManager.
            wcd_url (str | None): Required. The URL of the Weaviate Cloud Database instance used to save the tree.
                Defaults to the value of the `wcd_url` setting in the frontend config.
            wcd_api_key (str | None): Required. The API key for the Weaviate Cloud Database instance used to save the tree.
                Defaults to the value of the `wcd_api_key` setting in the frontend config.

        Returns:
            (dict): A dictionary whose keys are the conversation IDs and whose values are dictionaries containing the title and last update time of the tree.
                E.g.
                ```
                {
                    "12345678-XXX-YYYY-ZZZZ": {
                        "title": "Query Request",
                        "last_update_time": "2025-06-07T10:06:47.376000Z"
                    }
                }
                ```
        """
        local_user = await self.get_user_local(user_id)

        if wcd_url is None or wcd_api_key is None:
            save_location_client_manager = local_user[
                "frontend_config"
            ].save_location_client_manager
        else:
            save_location_client_manager = ClientManager(
                logger=logger,
                wcd_url=wcd_url,
                wcd_api_key=wcd_api_key,
            )

        return await get_saved_trees_weaviate(
            "ELYSIA_TREES__", save_location_client_manager, user_id
        )

    async def update_user_last_request(self, user_id: str):
        self.users[user_id]["last_request"] = datetime.datetime.now()

    def check_tree_timeout(self, user_id: str, conversation_id: str):
        if user_id not in self.users:
            return True
        elif conversation_id not in self.users[user_id]["tree_manager"].trees:
            return True
        return False

    async def process_tree(
        self,
        query: str,
        user_id: str,
        conversation_id: str,
        query_id: str,
        training_route: str = "",
        collection_names: list[str] = [],
        save_trees_to_weaviate: bool | None = None,
        wcd_url: str | None = None,
        wcd_api_key: str | None = None,
    ):
        """
        Wrapper for the TreeManager.process_tree() method.
        Which itself is a wrapper for the Tree.async_run() method.
        This is an async generator which yields results from the tree.async_run() method.
        Automatically sends error payloads if the user or tree has been timed out.

        Args:
            query (str): Required. The user input/prompt to process in the decision tree.
            user_id (str): Required. The unique identifier for the user.
            conversation_id (str): Required. The conversation ID which contains the tree.
                This should be the same conversation ID as the one used to initialise the tree (see `initialise_tree`).
            query_id (str): Required. A unique identifier for the query.
            training_route (str): Optional. The training route, a string of the form "tool1/tool2/tool1" etc.
                See the `tree.async_run()` method for more details.
            collection_names (list[str]): Optional. A list of collection names to use in the query.
                If not supplied, all collections will be used.
            save_trees_to_weaviate (bool | None): Optional. Whether to save the trees to a Weaviate instance,
                after the process_tree() method has finished.
                Defaults to the value of the `save_trees_to_weaviate` setting in the frontend config.
            wcd_url (str | None): Required. The URL of the Weaviate Cloud Database instance used to save the tree.
                Defaults to the value of the `wcd_url` setting in the frontend config.
            wcd_api_key (str | None): Required. The API key for the Weaviate Cloud Database instance used to save the tree.
                Defaults to the value of the `wcd_api_key` setting in the frontend config.
        """

        if self.check_user_timeout(user_id):
            user_timeout_error = UserTimeoutError()
            error_payload = await user_timeout_error.to_frontend(
                user_id, conversation_id, query_id
            )
            yield error_payload
            return

        if self.check_tree_timeout(user_id, conversation_id):
            if await self.check_tree_exists_weaviate(user_id, conversation_id):
                await self.load_tree(user_id, conversation_id)
            else:
                tree_timeout_error = TreeTimeoutError()
                error_payload = await tree_timeout_error.to_frontend(
                    user_id, conversation_id, query_id
                )
                yield error_payload
                return

        local_user = await self.get_user_local(user_id)
        await self.update_user_last_request(user_id)

        tree_manager: TreeManager = local_user["tree_manager"]

        async for yielded_result in tree_manager.process_tree(
            query,
            conversation_id,
            query_id,
            training_route,
            collection_names,
            local_user["client_manager"],
        ):
            yield yielded_result
            await self.update_user_last_request(user_id)

        if save_trees_to_weaviate is None:
            frontend_config: FrontendConfig = local_user["frontend_config"]
            save_trees_to_weaviate = frontend_config.config["save_trees_to_weaviate"]

        if save_trees_to_weaviate:
            await self.save_tree(user_id, conversation_id, wcd_url, wcd_api_key)



import os
from fastapi import APIRouter, Depends
from fastapi.responses import JSONResponse
from dotenv import load_dotenv
from uuid import uuid4

load_dotenv(override=True)

# API Types
from elysia.api.api_types import SaveConfigTreeData
from elysia.api.utils.config import Config
from elysia.api.core.log import logger
from elysia.api.dependencies.common import get_user_manager
from elysia.api.services.user import UserManager
from elysia.util.parsing import format_dict_to_serialisable
from elysia.tree.tree import Tree
from elysia.config import Settings, is_api_key

from weaviate.util import generate_uuid5


def rename_keys(config_item: dict):

    renamed_settings = {
        key.upper(): config_item["settings"][key] for key in config_item["settings"]
    }

    if "API_KEYS" in config_item["settings"]:
        renamed_api_keys = {
            key.lower(): config_item["settings"]["API_KEYS"][key]
            for key in config_item["settings"]["API_KEYS"]
        }
        renamed_settings["API_KEYS"] = renamed_api_keys

    renamed_config = {key.lower(): config_item[key] for key in config_item}
    renamed_config["settings"] = renamed_settings

    return renamed_config


router = APIRouter()


@router.get("/{user_id}/{conversation_id}")
async def get_tree_config(
    user_id: str,
    conversation_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Get the config for a single tree / conversation.
    """
    logger.debug(f"/get_tree_config API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Conversation ID: {conversation_id}")

    headers = {"Cache-Control": "no-cache"}

    try:
        tree: Tree = await user_manager.get_tree(user_id, conversation_id)
        config = Config(
            id="tree_config",
            name="Tree Config",
            settings=tree.settings,
            style=tree.tree_data.atlas.style,
            agent_description=tree.tree_data.atlas.agent_description,
            end_goal=tree.tree_data.atlas.end_goal,
            branch_initialisation=tree.branch_initialisation,
        )

    except Exception as e:
        logger.exception(f"Error in /get_tree_config API")
        return JSONResponse(content={"error": str(e), "config": {}}, headers=headers)

    return JSONResponse(
        content={"error": "", "config": config.to_json()}, headers=headers
    )


@router.post("/{user_id}/{conversation_id}/new")
async def new_tree_config(
    user_id: str,
    conversation_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Set the config in a single tree / conversation to the default values (gemini 2.0 flash or environment variables).
    Any changes will only affect this one tree / conversation.

    Args:
        user_id (str): The user ID.
        conversation_id (str): The conversation ID.
        user_manager (UserManager): The user manager instance.

    Returns:
        JSONResponse: A JSON response with the following attributes:
            error (str): An error message. Empty if no error.
            config (dict): A dictionary of the updated config values.
    """
    logger.debug(f"/new_tree_config API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Conversation ID: {conversation_id}")

    try:
        tree: Tree = await user_manager.get_tree(user_id, conversation_id)

        settings = Settings()
        settings.smart_setup()

        tree.configure(
            **{k: v for k, v in settings.to_json().items() if k != "API_KEYS"}
        )
        tree.configure(**settings.to_json()["API_KEYS"])

        tree.change_style("Informative, polite and friendly.")
        tree.change_agent_description(
            "You search and query Weaviate to satisfy the user's query, providing a concise summary of the results."
        )
        tree.change_end_goal(
            "You have satisfied the user's query, and provided a concise summary of the results. "
            "Or, you have exhausted all options available, or asked the user for clarification."
        )
        tree.set_branch_initialisation("one_branch")

        config = Config(
            id="tree_config",
            name="Tree Config",
            settings=tree.settings,
            style=tree.tree_data.atlas.style,
            agent_description=tree.tree_data.atlas.agent_description,
            end_goal=tree.tree_data.atlas.end_goal,
            branch_initialisation=tree.branch_initialisation,
        )

    except Exception as e:
        logger.exception(f"Error in /new_tree_config API")
        return JSONResponse(content={"error": str(e), "config": {}})

    return JSONResponse(content={"error": "", "config": config.to_json()})


@router.post("/{user_id}/{conversation_id}")
async def change_config_tree(
    user_id: str,
    conversation_id: str,
    data: SaveConfigTreeData,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Change the config for a single tree / conversation.
    The `config` dictionary can contain any of the settings (dict), style, agent_description, end_goal and branch_initialisation.
    You do not need to specify all the settings, only the ones you wish to change.

    Once you change the config, the tree will no longer inherit from the user's config.

    In this case, you cannot change the API keys, it will always inherit from the user's config.
    Any API keys provided will be ignored.

    Args:
        user_id (str): Required. The user ID.
        conversation_id (str): Required. The conversation ID.
        data (Config): Required. Containing the following attributes:
            settings(dict): Optional. A dictionary of config values to set, follows the same format as the Settings.configure() method.
            style (str): Optional. The writing style of the agent.
            agent_description (str): Optional. The description of the agent.
            end_goal (str): Optional. The end goal of the agent.
            branch_initialisation (str): Optional. The branch initialisation of the agent.
        user_manager (UserManager): The user manager instance.

    Returns:
        JSONResponse: A JSON response with the following attributes:
            error (str): An error message. Empty if no error.
            config (dict): A dictionary of the updated config values.
    """
    logger.debug(f"/change_config_tree API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Conversation ID: {conversation_id}")
    logger.debug(f"Settings: {data.settings}")
    logger.debug(f"Style: {data.style}")
    logger.debug(f"Agent description: {data.agent_description}")
    logger.debug(f"End goal: {data.end_goal}")
    logger.debug(f"Branch initialisation: {data.branch_initialisation}")

    try:
        tree: Tree = await user_manager.get_tree(user_id, conversation_id)

        if data.settings is not None:

            # tree level configs do not include API keys
            if "wcd_url" in data.settings:
                del data.settings["wcd_url"]

            api_keys = [key for key in data.settings if is_api_key(key)]
            for key in api_keys:
                del data.settings[key]

            tree.configure(**data.settings)

        if data.style is not None:
            tree.change_style(data.style)

        if data.agent_description is not None:
            tree.change_agent_description(data.agent_description)

        if data.end_goal is not None:
            tree.change_end_goal(data.end_goal)

        if data.branch_initialisation is not None:
            tree.set_branch_initialisation(data.branch_initialisation)

        config = Config(
            id="tree_config",
            name="Tree Config",
            settings=tree.settings,
            style=tree.tree_data.atlas.style,
            agent_description=tree.tree_data.atlas.agent_description,
            end_goal=tree.tree_data.atlas.end_goal,
            branch_initialisation=tree.branch_initialisation,
        )

    except Exception as e:
        logger.exception(f"Error in /change_config_tree API")
        return JSONResponse(content={"error": str(e), "config": {}})

    return JSONResponse(content={"error": "", "config": config.to_json()})


@router.post("/{user_id}/{conversation_id}/{config_id}/load")
async def load_config_tree(
    user_id: str,
    conversation_id: str,
    config_id: str,
    user_manager: UserManager = Depends(get_user_manager),
):
    """
    Load a config from the weaviate database for a single tree / conversation.
    This will set the config for the current tree.
    It will not affect other trees associated with the user.
    It will NOT update the ClientManager to use the new API keys.

    Args:
        user_id (str): The user ID.
        conversation_id (str): The conversation ID.
        config_id (str): The ID of the config to load. Can be found in the /list_configs API.
        user_manager (UserManager): The user manager instance.

    Returns:
        JSONResponse: A JSON response with the following attributes:
            error (str): An error message. Empty if no error.
            config (dict): A dictionary of the updated config values.
    """
    logger.debug(f"/load_config_tree API request received")
    logger.debug(f"User ID: {user_id}")
    logger.debug(f"Conversation ID: {conversation_id}")
    logger.debug(f"Config ID: {config_id}")

    try:
        # Retrieve the user info
        user = await user_manager.get_user_local(user_id)
        tree: Tree = await user_manager.get_tree(user_id, conversation_id)

        if (
            "frontend_config" not in user
            or "save_location_wcd_url" not in user["frontend_config"].__dict__
            or "save_location_wcd_api_key" not in user["frontend_config"].__dict__
        ):
            raise Exception("WCD URL or API key not found.")

        if (
            user["frontend_config"].save_location_wcd_url is None
            or user["frontend_config"].save_location_wcd_api_key is None
        ):
            raise Exception(
                "No valid destination for config load location found. "
                "Please update the save location using the /update_save_location API."
            )

        # Retrieve the config from the weaviate database
        async with user[
            "frontend_config"
        ].save_location_client_manager.connect_to_async_client() as client:
            uuid = generate_uuid5(config_id)
            collection = client.collections.get("ELYSIA_CONFIG__")
            config_item = await collection.query.fetch_object_by_id(uuid=uuid)

        # Load the configs to the current user
        format_dict_to_serialisable(config_item.properties)
        renamed_config = rename_keys(config_item.properties)

        tree.configure(
            **{
                c: renamed_config["settings"][c]
                for c in renamed_config["settings"]
                if (c != "api_keys" and not is_api_key(c))
            }
        )

        tree.change_style(renamed_config["style"])
        tree.change_agent_description(renamed_config["agent_description"])
        tree.change_end_goal(renamed_config["end_goal"])

        tree.set_branch_initialisation(renamed_config["branch_initialisation"])

    except Exception as e:
        logger.exception(f"Error in /load_config API")
        return JSONResponse(content={"error": str(e), "config": {}})

    return JSONResponse(content={"error": "", "config": renamed_config})



import dspy

from logging import Logger

from elysia.objects import Status, Tool
from elysia.tools.postprocessing.prompt_templates import ObjectSummaryPrompt
from elysia.tree.objects import TreeData
from elysia.util.client import ClientManager


class SummariseItems(Tool):
    def __init__(self, logger: Logger | None = None, **kwargs):
        super().__init__(
            name="query_postprocessing",
            description="""
            If the user has requested itemised summaries for retrieved objects,
            this tool summarises each object on an individual basis.
            """,
            status="",
            end=False,
        )

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        client_manager: ClientManager | None = None,
        **kwargs,
    ):

        if "items_to_summarise" in tree_data.environment.hidden_environment:

            # get objects from hidden environment
            objects_list = tree_data.environment.hidden_environment[
                "items_to_summarise"
            ]

            for obj in objects_list:
                object_summariser = dspy.ChainOfThought(ObjectSummaryPrompt)
                yield Status(
                    f"Summarising {len(obj.objects)} objects from {obj.metadata['collection_name']}..."
                )

                summariser = await object_summariser.aforward(
                    objects=obj.to_json(), lm=base_lm
                )

                obj.add_summaries(summariser.summaries)

                yield obj

            # remove items from hidden environment
            tree_data.environment.hidden_environment.pop("items_to_summarise")



import dspy
import pytest
from elysia.config import Settings, configure, reset_settings, IncorrectModelError
from elysia.tree.tree import Tree


class TestTreeConfig:

    def reset_global_settings(self):
        reset_settings()

    # def test_change_llm_in_tree_global(self):
    #     """
    #     When using global settings, is the change reflected in the tree?
    #     """
    #     self.reset_global_settings()

    #     # create a Tree (uses global settings by default)
    #     tree = Tree(low_memory=True)

    #     # should raise an error, no models set
    #     with pytest.raises(ValueError):
    #         tree.get_lm("base")
    #         tree.get_lm("complex")
    #     assert tree.base_lm is None
    #     assert tree.complex_lm is None

    #     # change the base model
    #     configure(base_model="gpt-4o-mini", base_provider="openai")

    #     # should now be changed (no error)
    #     base_lm_loaded_in_tree = tree.get_lm("base")  # should be a dspy.LM

    #     assert isinstance(base_lm_loaded_in_tree, dspy.LM)
    #     assert base_lm_loaded_in_tree.model == "openai/gpt-4o-mini"
    #     assert tree.settings.BASE_MODEL == "gpt-4o-mini"
    #     assert tree.settings.BASE_PROVIDER == "openai"

    #     # change the complex model
    #     configure(complex_model="gpt-4o", complex_provider="openai")

    #     # should now be changed (no error)
    #     complex_lm_loaded_in_tree = tree.get_lm("complex")

    #     assert isinstance(complex_lm_loaded_in_tree, dspy.LM)
    #     assert complex_lm_loaded_in_tree.model == "openai/gpt-4o"
    #     assert tree.settings.COMPLEX_MODEL == "gpt-4o"
    #     assert tree.settings.COMPLEX_PROVIDER == "openai"

    def test_change_llm_in_tree_local(self):
        """
        When using local settings, global settings should NOT be reflected in the tree
        """
        settings = Settings()

        # create a Tree with local settings obj
        tree = Tree(settings=settings, low_memory=True)

        # should raise an error, no models set
        with pytest.raises(IncorrectModelError):
            tree.base_lm

        with pytest.raises(IncorrectModelError):
            tree.complex_lm

        # change the models by global configure
        configure(
            base_model="gpt-4o-mini",
            base_provider="openai",
            complex_model="gpt-4o",
            complex_provider="openai",
        )

        # should still error because the tree doesnt use global configure
        with pytest.raises(IncorrectModelError):
            tree.base_lm

        with pytest.raises(IncorrectModelError):
            tree.complex_lm

        # change the models by local configure
        settings.configure(
            base_model="gpt-4o-mini",
            base_provider="openai",
            complex_model="gpt-4o",
            complex_provider="openai",
        )

        # should now be no errors
        base_lm_loaded_in_tree = tree.base_lm
        assert isinstance(base_lm_loaded_in_tree, dspy.LM)
        assert base_lm_loaded_in_tree.model == "openai/gpt-4o-mini"
        assert tree.settings.BASE_MODEL == "gpt-4o-mini"
        assert tree.settings.BASE_PROVIDER == "openai"

        complex_lm_loaded_in_tree = tree.complex_lm
        assert isinstance(complex_lm_loaded_in_tree, dspy.LM)
        assert complex_lm_loaded_in_tree.model == "openai/gpt-4o"
        assert tree.settings.COMPLEX_MODEL == "gpt-4o"
        assert tree.settings.COMPLEX_PROVIDER == "openai"

        self.reset_global_settings()

    def test_change_llm_with_debug(self):
        """
        When using debug, the models should be loaded into the Tree and not changed
        """
        settings = Settings()

        # create a Tree with local settings obj, should not error
        tree = Tree(settings=settings, low_memory=False)

        # change the models by global configure
        settings.configure(
            local=True,
            base_model="gpt-4o-mini",
            base_provider="openai",
            complex_model="gpt-4o",
            complex_provider="openai",
        )

        # should now be changed (no error)
        tree = Tree(settings=settings, low_memory=False)

        base_lm_loaded_in_tree = tree.base_lm
        assert isinstance(base_lm_loaded_in_tree, dspy.LM)
        assert base_lm_loaded_in_tree.model == "openai/gpt-4o-mini"
        assert tree.settings.BASE_MODEL == "gpt-4o-mini"
        assert tree.settings.BASE_PROVIDER == "openai"

        complex_lm_loaded_in_tree = tree.complex_lm
        assert isinstance(complex_lm_loaded_in_tree, dspy.LM)
        assert complex_lm_loaded_in_tree.model == "openai/gpt-4o"
        assert tree.settings.COMPLEX_MODEL == "gpt-4o"
        assert tree.settings.COMPLEX_PROVIDER == "openai"

        self.reset_global_settings()



import logging

from fastapi import FastAPI, Request
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from starlette.exceptions import HTTPException as StarletteHTTPException

logger = logging.getLogger(__name__)


def register_error_handlers(app: FastAPI):
    @app.exception_handler(StarletteHTTPException)
    async def http_exception_handler(request: Request, exc: StarletteHTTPException):
        logger.error(f"HTTP error occurred: {exc.detail}")
        return JSONResponse(
            status_code=exc.status_code,
            content={"detail": exc.detail},
        )

    @app.exception_handler(RequestValidationError)
    async def validation_exception_handler(
        request: Request, exc: RequestValidationError
    ):
        logger.error(f"Validation error: {exc.errors()}")
        return JSONResponse(
            status_code=422,
            content={"detail": exc.errors()},
        )

    @app.exception_handler(Exception)
    async def general_exception_handler(request: Request, exc: Exception):
        logger.error(f"Unexpected error: {str(exc)}")
        return JSONResponse(
            status_code=500,
            content={"detail": "Internal Server Error"},
        )



import pytest

from elysia.api.api_types import (
    InitialiseTreeData,
    QueryData,
    UpdateFrontendConfigData,
)
from elysia.api.routes.query import process
from elysia.api.dependencies.common import get_user_manager
from elysia.api.routes.init import initialise_user, initialise_tree
from elysia.api.routes.user_config import update_frontend_config
from elysia.api.routes.db import (
    save_tree,
    load_tree,
    get_saved_trees,
    delete_tree,
)

from fastapi.responses import JSONResponse
import json


def read_response(response: JSONResponse):
    return json.loads(response.body)


from elysia.api.core.log import logger, set_log_level

set_log_level("CRITICAL")


def read_response(response: JSONResponse):
    return json.loads(response.body)


class fake_websocket:
    results = []

    async def send_json(self, data: dict):
        self.results.append(data)


async def initialise_user_and_tree(user_id: str, conversation_id: str):
    user_manager = get_user_manager()

    response = await initialise_user(
        user_id,
        user_manager,
    )

    response = await initialise_tree(
        user_id,
        conversation_id,
        InitialiseTreeData(
            low_memory=False,
        ),
        user_manager,
    )


async def delete_user_and_tree(user_id: str, conversation_id: str):
    user_manager = get_user_manager()
    await user_manager.delete_tree(user_id, conversation_id)


class TestSaveLoad:

    @pytest.mark.asyncio
    async def test_manual_save_load_cycle(self):

        user_id = "test_user_full_save_load_cycle"
        conversation_id = "test_conversation_full_save_load_cycle"
        query_id = "test_query_full_save_load_cycle"

        websocket = fake_websocket()
        user_manager = get_user_manager()

        await initialise_user_and_tree(user_id, conversation_id)

        try:
            await update_frontend_config(
                user_id=user_id,
                data=UpdateFrontendConfigData(
                    config={
                        "save_trees_to_weaviate": False,
                    },
                ),
                user_manager=user_manager,
            )

            await process(
                QueryData(
                    user_id=user_id,
                    conversation_id=conversation_id,
                    query="hi",
                    query_id=query_id,
                    collection_names=[],
                ).model_dump(),
                websocket,
                user_manager,
            )

            # check that the tree is not saved to weaviate (automatic saving turned off)
            response = await get_saved_trees(
                user_id=user_id,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""
            assert conversation_id not in list(read_response(response)["trees"].keys())

            response = await save_tree(
                user_id=user_id,
                conversation_id=conversation_id,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""

            response = await load_tree(
                user_id=user_id,
                conversation_id=conversation_id,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""
            assert read_response(response)["rebuild"] != []
            for rebuilt in read_response(response)["rebuild"]:
                assert "type" in rebuilt
                assert "conversation_id" in rebuilt
                assert rebuilt["conversation_id"] == conversation_id
                assert "id" in rebuilt
                if "user_id" in rebuilt:
                    assert rebuilt["user_id"] == user_id

            response = await get_saved_trees(
                user_id=user_id,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""
            assert len(list(read_response(response)["trees"].keys())) >= 1
            old_len = len(list(read_response(response)["trees"].keys()))
            assert "test_conversation_full_save_load_cycle" in list(
                read_response(response)["trees"].keys()
            )

            response = await delete_tree(
                user_id=user_id,
                conversation_id=conversation_id,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""

            response = await get_saved_trees(
                user_id=user_id,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""
            assert len(list(read_response(response)["trees"].keys())) < old_len
            assert "test_conversation_full_save_load_cycle" not in list(
                read_response(response)["trees"].keys()
            )
        finally:
            await delete_user_and_tree(user_id, conversation_id)

    @pytest.mark.asyncio
    async def test_automatic_save_load_cycle(self):

        user_id = "test_user_automatic_save_load_cycle"
        conversation_id = "test_conversation_automatic_save_load_cycle"
        query_id = "test_query_automatic_save_load_cycle"

        websocket = fake_websocket()
        user_manager = get_user_manager()

        await initialise_user_and_tree(user_id, conversation_id)

        try:
            await update_frontend_config(
                user_id=user_id,
                data=UpdateFrontendConfigData(
                    config={
                        "save_trees_to_weaviate": True,
                    },
                ),
                user_manager=user_manager,
            )

            await process(
                QueryData(
                    user_id=user_id,
                    conversation_id=conversation_id,
                    query="hi",
                    query_id=query_id,
                    collection_names=[],
                ).model_dump(),
                websocket,
                user_manager,
            )

            response = await get_saved_trees(
                user_id=user_id,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""
            assert len(list(read_response(response)["trees"].keys())) >= 1
            old_len = len(list(read_response(response)["trees"].keys()))
            assert "test_conversation_automatic_save_load_cycle" in list(
                read_response(response)["trees"].keys()
            )

            response = await load_tree(
                user_id=user_id,
                conversation_id=conversation_id,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""
            assert read_response(response)["rebuild"] != []
            for rebuilt in read_response(response)["rebuild"]:
                assert "type" in rebuilt
                assert "conversation_id" in rebuilt
                assert rebuilt["conversation_id"] == conversation_id
                assert "id" in rebuilt
                if "user_id" in rebuilt:
                    assert rebuilt["user_id"] == user_id

            response = await delete_tree(
                user_id=user_id,
                conversation_id=conversation_id,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""

            response = await get_saved_trees(
                user_id=user_id,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""
            assert len(list(read_response(response)["trees"].keys())) < old_len
            assert "test_conversation_automatic_save_load_cycle" not in list(
                read_response(response)["trees"].keys()
            )
        finally:
            await delete_user_and_tree(user_id, conversation_id)

    @pytest.mark.asyncio
    async def test_automatic_load_on_tree_timeout(self):

        user_id = "test_user_automatic_load_on_tree_timeout"
        conversation_id = "test_conversation_automatic_load_on_tree_timeout"
        query_id = "test_query_automatic_load_on_tree_timeout"

        websocket = fake_websocket()
        user_manager = get_user_manager()

        await initialise_user_and_tree(user_id, conversation_id)

        try:

            # do a query
            await process(
                QueryData(
                    user_id=user_id,
                    conversation_id=conversation_id,
                    query="hi",
                    query_id=query_id,
                    collection_names=[],
                ).model_dump(),
                websocket,
                user_manager,
            )

            # store number of outputs in websocket results
            num_outputs = len(websocket.results)

            # the tree should be saved to weaviate
            response = await get_saved_trees(
                user_id=user_id,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""
            assert len(list(read_response(response)["trees"].keys())) >= 1

            # timeout the tree manually by deleting it from the treemanager
            local_user = await user_manager.get_user_local(user_id)
            tree_manager = local_user["tree_manager"]
            del tree_manager.trees[conversation_id]

            # check the tree exists in weaviate
            tree_exists_response = await user_manager.check_tree_exists_weaviate(
                user_id, conversation_id
            )
            assert tree_exists_response

            # do a query - this should not error
            await process(
                QueryData(
                    user_id=user_id,
                    conversation_id=conversation_id,
                    query="hi again",
                    query_id=query_id,
                    collection_names=[],
                ).model_dump(),
                websocket,
                user_manager,
            )
            assert len(websocket.results) > num_outputs
            assert "type" in websocket.results[-1]
            assert websocket.results[-1]["type"] != "tree_timeout_error"
        finally:
            await delete_user_and_tree(user_id, conversation_id)

    @pytest.mark.asyncio
    async def test_saved_trees_with_different_users(self):

        user_id = "test_user_saved_trees_with_different_users_1"
        conversation_id = "test_conversation_saved_trees_with_different_users_1"
        query_id = "test_query_saved_trees_with_different_users_1"

        user_id_2 = "test_user_saved_trees_with_different_users_2"
        conversation_id_2 = "test_conversation_saved_trees_with_different_users_2"
        query_id_2 = "test_query_saved_trees_with_different_users_2"

        websocket = fake_websocket()
        user_manager = get_user_manager()

        await initialise_user_and_tree(user_id, conversation_id)
        await initialise_user_and_tree(user_id_2, conversation_id_2)
        try:

            # do a query with this user, automatically saves
            await process(
                QueryData(
                    user_id=user_id,
                    conversation_id=conversation_id,
                    query="hi",
                    query_id=query_id,
                    collection_names=[],
                ).model_dump(),
                websocket,
                user_manager,
            )

            # do a query with this user, automatically saves
            await process(
                QueryData(
                    user_id=user_id_2,
                    conversation_id=conversation_id_2,
                    query="hi",
                    query_id=query_id_2,
                    collection_names=[],
                ).model_dump(),
                websocket,
                user_manager,
            )

            # check that the trees exist in two separate places
            response = await get_saved_trees(
                user_id=user_id,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""
            assert conversation_id in list(read_response(response)["trees"].keys())
            assert conversation_id_2 not in list(
                read_response(response)["trees"].keys()
            )

            # check that the trees exist in two separate places
            response = await get_saved_trees(
                user_id=user_id_2,
                user_manager=user_manager,
            )
            assert read_response(response)["error"] == ""
            assert conversation_id not in list(read_response(response)["trees"].keys())
            assert conversation_id_2 in list(read_response(response)["trees"].keys())
        finally:
            await delete_user_and_tree(user_id, conversation_id)
            await delete_user_and_tree(user_id_2, conversation_id_2)



from elysia.util.client import ClientManager
import json
import dspy
import random
from weaviate.classes.query import Filter, MetadataQuery


async def retrieve_feedback(
    client_manager: ClientManager, user_prompt: str, model: str, n: int = 6
):
    """
    Retrieve similar examples from the database.
    """

    # semantic search for similar examples
    async with client_manager.connect_to_async_client() as client:
        if not await client.collections.exists("ELYSIA_FEEDBACK__"):
            return [], []

        feedback_collection = client.collections.get("ELYSIA_FEEDBACK__")

        # find superpositive examples
        superpositive_feedback = await feedback_collection.query.near_text(
            query=user_prompt,
            filters=Filter.all_of(
                [
                    Filter.by_property("modules_used").contains_any([model]),
                    Filter.by_property("feedback").equal(2.0),
                ]
            ),
            certainty=0.7,
            limit=n,
            return_metadata=MetadataQuery(distance=True, certainty=True),
        )

        if len(superpositive_feedback.objects) < n:

            # find positive examples
            positive_feedback = await feedback_collection.query.near_text(
                query=user_prompt,
                filters=Filter.all_of(
                    [
                        Filter.by_property("modules_used").contains_any([model]),
                        Filter.by_property("feedback").equal(1.0),
                    ],
                ),
                certainty=0.7,
                limit=n,
                return_metadata=MetadataQuery(distance=True, certainty=True),
            )

            feedback_objects = superpositive_feedback.objects
            feedback_objects.extend(
                positive_feedback.objects[: (n - len(superpositive_feedback.objects))]
            )

        else:
            feedback_objects = superpositive_feedback.objects

    # get training updates
    training_updates = [
        json.loads(f.properties["training_updates"])  # type: ignore
        for f in feedback_objects
    ]
    uuids = [str(f.uuid) for f in feedback_objects]

    relevant_updates = []
    relevant_uuids = []
    for i, update in enumerate(training_updates):
        for inner_update in update:
            if inner_update["module_name"] == model:
                relevant_updates.append(inner_update)
        relevant_uuids.append(uuids[i])

    # take max n randomly selected updates
    random.shuffle(relevant_updates)
    relevant_updates = relevant_updates[:n]

    examples = []
    for update in relevant_updates:
        examples.append(
            dspy.Example(
                {
                    **{k: v for k, v in update["inputs"].items()},
                    **{k: v for k, v in update["outputs"].items()},
                }
            ).with_inputs(
                *update["inputs"].keys(),
            )
        )

    return examples, relevant_uuids



import asyncio
import time

import psutil
from typing import Callable
from fastapi import WebSocket
from starlette.websockets import WebSocketDisconnect

# Logging
from elysia.api.core.log import logger

# Objects
from elysia.api.utils.default_payloads import error_payload


async def help_websocket(websocket: WebSocket, ws_route: Callable):
    memory_process = psutil.Process()
    # initial_memory = memory_process.memory_info().rss
    try:
        await websocket.accept()
        while True:
            try:
                # start_time = time.time()
                # Wait for a message from the client
                # logger.info(f"Memory usage before receiving: {psutil.Process().memory_info().rss / 1024 / 1024}MB")
                try:
                    data = None
                    last_communication = time.time()

                    # Custom timeout handler
                    while True:
                        if time.time() - last_communication > 60:
                            # logger.warning("No communication for 60 seconds - sending heartbeat")
                            await websocket.send_json({"type": "heartbeat"})
                            last_communication = time.time()

                        try:
                            # Use a short timeout for receive_json to allow checking the timer
                            data = await asyncio.wait_for(
                                websocket.receive_json(), timeout=1.0
                            )
                            last_communication = time.time()

                            # Process the received data
                            await ws_route(data, websocket)
                            last_communication = (
                                time.time()
                            )  # Update timer after processing

                            # Check if it's a disconnect request
                            if data.get("type") == "disconnect":
                                return

                        except asyncio.TimeoutError:
                            continue

                except Exception as e:
                    error = error_payload(text=str(e), conversation_id="", query_id="")
                    await websocket.send_json(error)
                    logger.error(f"Error in websocket communication: {str(e)}")

                # logger.info(f"Memory usage after receiving: {psutil.Process().memory_info().rss / 1024 / 1024}MB")
                # logger.info(f"Processing time: {time.time() - start_time}s")

            except WebSocketDisconnect:
                # logger.info("WebSocket disconnected", exc_info=True)
                break  # Exit the loop on disconnect

            except RuntimeError as e:
                if (
                    "Cannot call 'receive' once a disconnect message has been received"
                    in str(e)
                ):
                    # logger.info("WebSocket already disconnected")
                    break  # Exit the loop if the connection is already closed
                else:
                    raise  # Re-raise other RuntimeErrors

            except Exception as e:
                logger.error(f"Error in WebSocket: {str(e)}")
                try:
                    if data and "conversation_id" in data and "query_id" in data:
                        error = error_payload(
                            text=str(e),
                            conversation_id=data["conversation_id"],
                            query_id=data["query_id"],
                        )
                        await websocket.send_json(error)
                    elif data and "conversation_id" in data:
                        error = error_payload(
                            text=str(e),
                            conversation_id=data["conversation_id"],
                            query_id="",
                        )
                        await websocket.send_json(error)
                    else:
                        error = error_payload(
                            text=str(e), conversation_id="", query_id=""
                        )
                        await websocket.send_json(error)

                except RuntimeError:
                    logger.warning(
                        "Failed to send error message, WebSocket might be closed"
                    )
                break  # Exit the loop after sending the error message

    except Exception as e:
        logger.warning(f"Closing WebSocket: {str(e)}")
    finally:
        try:
            await websocket.close()
        except RuntimeError:
            logger.info("WebSocket already closed")



import os
from contextlib import asynccontextmanager

from apscheduler.schedulers.asyncio import AsyncIOScheduler

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from fastapi.responses import FileResponse

from elysia.api.core.log import logger, set_log_level
from elysia.api.dependencies.common import get_user_manager
from elysia.api.middleware.error_handlers import register_error_handlers
from elysia.api.routes import (
    collections,
    feedback,
    init,
    processor,
    query,
    user_config,
    tree_config,
    utils,
    tools,
    db,
)
from elysia.api.services.user import UserManager
from elysia.api.utils.resources import print_resources


from pathlib import Path


async def check_timeouts():
    user_manager = get_user_manager()
    await user_manager.check_all_trees_timeout()


async def output_resources():
    user_manager = get_user_manager()
    await print_resources(user_manager, save_to_file=True)


async def check_restart_clients():
    user_manager = get_user_manager()
    await user_manager.check_restart_clients()


@asynccontextmanager
async def lifespan(app: FastAPI):
    user_manager = get_user_manager()

    scheduler = AsyncIOScheduler()
    set_log_level("INFO")

    # use prime numbers for intervals so they don't overlap
    scheduler.add_job(check_timeouts, "interval", seconds=29)
    scheduler.add_job(check_restart_clients, "interval", seconds=31)
    scheduler.add_job(output_resources, "interval", seconds=1103)

    scheduler.start()
    yield
    scheduler.shutdown()

    await user_manager.close_all_clients()


# Create FastAPI app instance
app = FastAPI(title="Elysia API", version="0.3.0", lifespan=lifespan)

# Add CORS middleware
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Register error handlers
register_error_handlers(app)

# Include routers
app.include_router(init.router, prefix="/init", tags=["init"])
app.include_router(query.router, prefix="/ws", tags=["websockets"])
app.include_router(processor.router, prefix="/ws", tags=["websockets"])
app.include_router(collections.router, prefix="/collections", tags=["collections"])
app.include_router(user_config.router, prefix="/user/config", tags=["user config"])
app.include_router(tree_config.router, prefix="/tree/config", tags=["tree config"])
app.include_router(feedback.router, prefix="/feedback", tags=["feedback"])
app.include_router(utils.router, prefix="/util", tags=["utilities"])
app.include_router(tools.router, prefix="/tools", tags=["tools"])
app.include_router(db.router, prefix="/db", tags=["db"])


# Health check endpoint (kept in main app.py due to its simplicity)
@app.get("/api/health", tags=["health"])
async def health_check():
    """Health check endpoint."""
    logger.info("Health check requested")
    return {"status": "healthy"}


# Mount the app from static files
BASE_DIR = Path(__file__).resolve().parent

# Serve NextJS _next assets at root level (this is crucial!)
app.mount(
    "/_next",
    StaticFiles(directory=BASE_DIR / "static/_next"),
    name="next-assets",
)

# Serve other static files
app.mount("/static", StaticFiles(directory=BASE_DIR / "static"), name="app")


@app.get("/")
@app.head("/")
async def serve_frontend():
    if os.path.exists(os.path.join(BASE_DIR, "static/index.html")):
        return FileResponse(os.path.join(BASE_DIR, "static/index.html"))
    else:
        return None



from elysia.tools.visualisation.objects import (
    BarChart,
    HistogramChart,
    ScatterOrLineChart,
)
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from datetime import datetime
import numpy as np


def convert_chart_types_to_matplotlib(
    charts: list[BarChart | HistogramChart | ScatterOrLineChart],
) -> Figure:
    plt.style.use("seaborn-v0_8-whitegrid")
    n_charts = len(charts)
    n_cols = min(2, n_charts)  # Max 2 columns
    n_rows = (n_charts + n_cols - 1) // n_cols  # Ceiling division

    # Calculate figure size based on number of charts while respecting max size
    fig_width = min(18, 7 * n_cols)
    fig_height = min(15, 5 * n_rows)

    fig, axs = plt.subplots(
        n_rows, n_cols, figsize=(fig_width, fig_height), squeeze=False
    )

    # colour cycle
    colours = plt.rcParams["axes.prop_cycle"].by_key()["color"]

    for chart, ax in zip(charts, axs.flatten()):

        if isinstance(chart, BarChart):
            _create_bar_chart(chart, ax, colours)
        elif isinstance(chart, HistogramChart):
            _create_histogram(chart, ax, colours)
        elif isinstance(chart, ScatterOrLineChart):
            _create_scatter_or_line(chart, ax, colours)

        ax.set_title(chart.title)

    # Hide empty subplots
    for i in range(len(charts), len(axs.flatten())):
        axs.flatten()[i].set_visible(False)

    fig.tight_layout()
    return fig


def _create_scatter_or_line(chart: ScatterOrLineChart, ax, colours):

    x_data_points = [dp.value for dp in chart.data.x_axis]
    num_labels = 0
    for i, y_data in enumerate(chart.data.y_axis):

        y_data_points = [d.value for d in y_data.data_points]
        y_data_labels = [d.label for d in y_data.data_points]
        y_data_kind = y_data.kind

        colour = colours[i % len(colours)]

        if (
            chart.data.normalize_y_axis
            and not any(isinstance(dp, str) for dp in y_data_points)
            and not any(isinstance(dp, datetime) for dp in y_data_points)
        ):
            y_data_points = [dp / max(y_data_points) for dp in y_data_points]  # type: ignore

        if y_data_kind == "scatter":
            ax.scatter(x_data_points, y_data_points, label=y_data.label, color=colour)

        elif y_data_kind == "line":
            ax.plot(x_data_points, y_data_points, label=y_data.label, color=colour)

        for j, y_data_label in enumerate(y_data_labels):
            if y_data_label != "":
                ax.scatter(x_data_points[j], y_data_points[j], color=colour)
                if num_labels >= 1 and chart.data.normalize_y_axis:
                    xytext = (0, 10 + 5 * num_labels)
                else:
                    xytext = (0, 10)

                ax.annotate(
                    y_data_label,
                    (x_data_points[j], y_data_points[j]),
                    textcoords="offset pixels",
                    xytext=xytext,
                    ha="center",
                    va="center",
                    bbox=dict(facecolor="white", edgecolor="none", alpha=0.7, pad=2),
                )
                num_labels += 1

    ax.set_xlabel(chart.x_axis_label)

    # Rotate x-axis labels to prevent overlap
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right")

    if chart.data.normalize_y_axis and (
        "normalized" not in chart.y_axis_label.lower()
        and "normalised" not in chart.y_axis_label.lower()
    ):
        ax.set_ylabel(f"{chart.y_axis_label} (Normalized)")
    else:
        ax.set_ylabel(chart.y_axis_label)

    ax.set_title(chart.title)
    if len(chart.data.y_axis) > 1:
        ax.legend(bbox_to_anchor=(1.05, 1), loc="upper left")


def _create_bar_chart(chart: BarChart, ax, colours):
    """Create a bar chart on the given axis."""
    width = 0.8 / len(chart.data.y_values)

    x = np.arange(len(chart.data.x_labels))

    for i, (label, values) in enumerate(chart.data.y_values.items()):
        offset = width * i
        ax.bar(
            x + offset,
            values,
            color=colours[i % len(colours)],
            width=width,
            label=label,
            align="center",
        )

    ax.set_xticks(
        x + width * (len(chart.data.y_values) / 2) * 0.75, chart.data.x_labels
    )
    ax.set_xlabel(chart.x_axis_label)
    ax.set_ylabel(chart.y_axis_label)
    ax.set_title(chart.title)
    if len(chart.data.y_values.keys()) > 1:
        ax.legend(bbox_to_anchor=(1.05, 1))


def _create_histogram(chart: HistogramChart, ax, colours):
    """Create a histogram on the given axis."""

    for i, (label, values) in enumerate(chart.data.items()):
        ax.hist(
            values.distribution, label=label, color=colours[i % len(colours)], alpha=0.7
        )

    if len(chart.data.keys()) == 1:
        ax.set_xlabel(list(chart.data.keys())[0])
    else:
        ax.set_xlabel("Values")

    ax.set_ylabel("Density")
    ax.set_title(chart.title)
    ax.legend()



import os
from re import S
import pytest
import asyncio
import dspy
from dspy import LM
from elysia.objects import Result
from copy import deepcopy
from elysia import Tool
from elysia.tree.objects import TreeData
from elysia.objects import Response, tool
from elysia.config import Settings, configure
from elysia.tree.tree import Tree
from elysia.tools.text.text import TextResponse
from elysia.util.client import ClientManager
from elysia.tools.retrieval.query import Query
from elysia.tools.retrieval.aggregate import Aggregate

configure(logging_level="CRITICAL")


# == define some tools ==
class CheckResult(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="check_result",
            description="Check the result of the previous tool.",
        )

    async def __call__(
        self, tree_data, inputs, base_lm, complex_lm, client_manager, **kwargs
    ):
        yield Response("Looks good to me!")


class SendEmail(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="send_email",
            description="Send an email.",
            inputs={
                "email_address": {
                    "type": str,
                    "description": "The email address to send the email to.",
                    "required": True,
                }
            },
        )

    async def __call__(
        self, tree_data, inputs, base_lm, complex_lm, client_manager, **kwargs
    ):
        yield Response(f"Email sent to {inputs['email_address']}!")


class RunIfTrueFalseTool(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="run_if_true_false_tool",
            description="Always returns False",
        )

    async def is_tool_available(self, tree_data, base_lm, complex_lm, client_manager):
        return False

    async def run_if_true(
        self,
        tree_data: dict,
        base_lm: LM,
        complex_lm: LM,
        client_manager: ClientManager,
    ):
        return False, {}

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: LM,
        complex_lm: LM,
        client_manager: ClientManager,
        **kwargs,
    ):

        tree_data.environment.add(
            "rule_tool",
            Result(
                objects=[{"message": "Rule tool called!!!"}],
                payload_type="text",
                name="rule_tool",
            ),
        )
        yield False


class RunIfTrueTrueTool(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="run_if_true_true_tool",
            description="Always returns True",
        )

    async def is_tool_available(self, tree_data, base_lm, complex_lm, client_manager):
        return False

    async def run_if_true(
        self,
        tree_data: TreeData,
        base_lm: LM,
        complex_lm: LM,
        client_manager: ClientManager,
    ):
        return True, {}

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: LM,
        complex_lm: LM,
        client_manager: ClientManager,
        **kwargs,
    ):

        tree_data.environment.add(
            "rule_tool",
            Result(
                objects=[{"message": "Rule tool called!!!"}],
                payload_type="text",
                name="rule_tool",
            ),
        )
        yield True


class RunIfTrueTrueWithInputsTool(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="run_if_true_true_tool",
            description="Always returns True",
            inputs={
                "message": {
                    "type": "string",
                    "description": "A test input",
                    "default": "This is the default input",
                }
            },
        )
        self.use_default_inputs = kwargs.get("use_default_inputs", False)

    async def is_tool_available(self, tree_data, base_lm, complex_lm, client_manager):
        return False

    async def run_if_true(
        self,
        tree_data: TreeData,
        base_lm: LM,
        complex_lm: LM,
        client_manager: ClientManager,
    ):
        if self.use_default_inputs:
            return True, {}
        else:
            return True, {"message": "This is not the default input"}

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: LM,
        complex_lm: LM,
        client_manager: ClientManager,
        **kwargs,
    ):

        tree_data.environment.add(
            "rule_tool",
            Result(
                objects=[{"message": inputs["message"]}],
                payload_type="text",
                name="rule_tool",
            ),
        )
        yield True


class ToolNotAvailable(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="always_pick_this_tool",
            description="No matter what, always pick this tool, and END the conversation.",
            end=True,
        )

    async def is_tool_available(self, tree_data, base_lm, complex_lm, client_manager):
        return False

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: LM,
        complex_lm: LM,
        client_manager: ClientManager,
        **kwargs,
    ):
        yield True


class ToolAvailable(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="always_pick_this_tool",
            description="No matter what, always pick this tool, and END the conversation.",
            end=True,
        )

    async def is_tool_available(self, tree_data, base_lm, complex_lm, client_manager):
        return True

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: LM,
        complex_lm: LM,
        client_manager: ClientManager,
        **kwargs,
    ):
        yield True


class IncorrectToolInitialisation_kwargs_init(Tool):
    def __init__(self):
        super().__init__(
            name="incorrect_tool_initialisation",
            description="This tool should not be initialised",
        )


class IncorrectToolInitialisation_kwargs_call(Tool):
    def __init__(self):
        super().__init__(
            name="incorrect_tool_initialisation",
            description="This tool should not be initialised",
        )

    async def __call__(self, tree_data, inputs, base_lm, complex_lm, client_manager):
        yield True


class IncorrectToolInitialisation_call_non_async(Tool):
    def __init__(self):
        super().__init__(
            name="incorrect_tool_initialisation",
            description="This tool should not be initialised",
        )

    def __call__(self, tree_data, inputs, base_lm, complex_lm, client_manager):
        return True


class IncorrectToolInitialisation_call_non_async_generator(Tool):
    def __init__(self):
        super().__init__(
            name="incorrect_tool_initialisation",
            description="This tool should not be initialised",
        )

    async def __call__(self, tree_data, inputs, base_lm, complex_lm, client_manager):
        return True


@tool
async def example_async_decorator_tool():
    return "This is a test response from the decorator tool"


async def run_tree(
    user_prompt: str,
    collection_names: list[str],
    tools: list[Tool],
    remove_tools: bool = False,
    **kwargs,
):

    settings = Settings()
    settings.configure(
        base_model="gpt-4o-mini",
        base_provider="openai",
        complex_model="gpt-4o",
        complex_provider="openai",
    )

    tree = Tree(
        low_memory=False,
        branch_initialisation="empty",
        settings=settings,
    )

    if not remove_tools:
        tree.add_tool(TextResponse, root=True)

    for tool in tools:
        tree.add_tool(tool, root=True, **kwargs)

    async for result in tree.async_run(
        user_prompt,
        collection_names=collection_names,
    ):
        pass

    return tree


@pytest.mark.asyncio
async def test_run_if_true_false_tool():

    tree = await run_tree("Hello", [], [RunIfTrueFalseTool])

    assert "rule_tool" not in tree.tree_data.environment.environment


@pytest.mark.asyncio
async def test_run_if_true_true_tool():

    tree = await run_tree("Hello", [], [RunIfTrueTrueTool])

    assert "rule_tool" in tree.tree_data.environment.environment


@pytest.mark.asyncio
async def test_run_if_true_true_with_default_inputs_tool():
    tree = await run_tree(
        "Hello",
        [],
        [RunIfTrueTrueWithInputsTool],
        use_default_inputs=True,
    )

    assert "rule_tool" in tree.tree_data.environment.environment
    assert (
        tree.tree_data.environment.environment["rule_tool"]["rule_tool"][0]["objects"][
            0
        ]["message"]
        == "This is the default input"
    )


@pytest.mark.asyncio
async def test_run_if_true_true_with_non_default_inputs_tool():
    tree = await run_tree(
        "Hello",
        [],
        [RunIfTrueTrueWithInputsTool],
        use_default_inputs=False,
    )

    assert "rule_tool" in tree.tree_data.environment.environment
    assert (
        tree.tree_data.environment.environment["rule_tool"]["rule_tool"][0]["objects"][
            0
        ]["message"]
        == "This is not the default input"
    )


@pytest.mark.asyncio
async def test_tool_not_available():

    with pytest.raises(ValueError):  # should have no tools available
        await run_tree("Hello", [], [ToolNotAvailable], remove_tools=True)


@pytest.mark.asyncio
async def test_tool_available():

    tree = await run_tree("Hello", [], [ToolAvailable], remove_tools=True)
    all_decision_history = []
    for iteration in tree.decision_history:
        all_decision_history.extend(iteration)

    assert "always_pick_this_tool" in all_decision_history


def test_incorrect_tool_initialisation():
    with pytest.raises(TypeError):
        tree = Tree(
            low_memory=False,
            branch_initialisation="empty",
            settings=Settings.from_smart_setup(),
        )
        tree.add_tool(IncorrectToolInitialisation_kwargs_init)

    with pytest.raises(TypeError):
        tree = Tree(
            low_memory=False,
            branch_initialisation="empty",
            settings=Settings.from_smart_setup(),
        )
        tree.add_tool(IncorrectToolInitialisation_kwargs_call)

    with pytest.raises(TypeError):
        tree = Tree(
            low_memory=False,
            branch_initialisation="empty",
            settings=Settings.from_smart_setup(),
        )
        tree.add_tool(IncorrectToolInitialisation_call_non_async)

    with pytest.raises(TypeError):
        tree = Tree(
            low_memory=False,
            branch_initialisation="empty",
            settings=Settings.from_smart_setup(),
        )
        tree.add_tool(IncorrectToolInitialisation_call_non_async_generator)


@pytest.mark.asyncio
async def test_example_decorator_tool():

    # Async tools should work
    tree = await run_tree(
        "Hello", [], [example_async_decorator_tool], remove_tools=True
    )
    assert "example_async_decorator_tool" in tree.tools

    # Sync tools should not work
    with pytest.raises(TypeError):

        @tool
        def example_sync_decorator_tool():
            return "This is a test response from the decorator tool"


@pytest.mark.asyncio
async def test_example_decorator_tool_from_tree():

    # Async tools should work
    tree = Tree()

    @tool(tree=tree)
    async def example_async_decorator_tool_from_tree():
        return "This is a test response from the decorator tool"

    assert "example_async_decorator_tool_from_tree" in tree.tools


@pytest.mark.asyncio
async def test_add_tool_with_stem_tool():
    tree = Tree(
        low_memory=False,
        branch_initialisation="empty",
        settings=Settings.from_smart_setup(),
    )

    tree.add_branch(
        branch_id="search",
        instruction="Search for information",
        description="Search for information",
        root=False,
        from_branch_id="base",
    )

    tree.add_tool(Query, branch_id="search")
    tree.add_tool(Aggregate, branch_id="search")

    # no query in base branch
    with pytest.raises(ValueError):
        tree.add_tool(CheckResult, branch_id="base", from_tool_ids=["query"])

    # random_text is not a tool
    with pytest.raises(ValueError):
        tree.add_tool(CheckResult, branch_id="base", from_tool_ids=["random_text"])

    # query is in search branch, should not error
    tree.add_tool(CheckResult, branch_id="search", from_tool_ids=["query"])
    assert "check_result" in tree.tools

    # query (from search) should now be a decision node
    assert "search.query" in tree.decision_nodes

    # must specify all from_tool_ids or it will error
    with pytest.raises(ValueError):
        tree.add_tool(SendEmail, branch_id="search", from_tool_ids=["check_result"])

    # correct usage, should not error
    tree.add_tool(
        SendEmail, branch_id="search", from_tool_ids=["query", "check_result"]
    )

    assert "send_email" in tree.tools
    assert "search.query.check_result" in tree.decision_nodes

    # check the tree.tree is correct
    assert (
        "check_result" in tree.tree["options"]["search"]["options"]["query"]["options"]
    )
    assert (
        "send_email"
        in tree.tree["options"]["search"]["options"]["query"]["options"][
            "check_result"
        ]["options"]
    )

    # remove with the wrong from_tool_ids
    with pytest.raises(ValueError):
        tree.remove_tool(
            tool_name="send_email",
            branch_id="search",
            from_tool_ids=["query"],
        )

    # remove from the wrong branch
    with pytest.raises(ValueError):
        tree.remove_tool(
            tool_name="send_email",
            branch_id="base",
            from_tool_ids=["query", "check_result"],
        )

    # remove with the correct from_tool_ids
    tree.remove_tool(
        tool_name="send_email",
        branch_id="search",
        from_tool_ids=["query", "check_result"],
    )

    assert "send_email" not in tree.tools
    assert "search.query.check_result" not in tree.decision_nodes

    # check the tree.tree is correct
    assert (
        "send_email"
        not in tree.tree["options"]["search"]["options"]["query"]["options"][
            "check_result"
        ]["options"]
    )

    # add the tool back in
    tree.add_tool(
        SendEmail, branch_id="search", from_tool_ids=["query", "check_result"]
    )

    # remove the tool that this tool stems from
    tree.remove_tool(
        tool_name="check_result",
        branch_id="search",
        from_tool_ids=["query"],
    )

    # check that the tool is removed
    assert "check_result" not in tree.tools
    assert "search.query" not in tree.decision_nodes

    # check that the stemmed tool is removed from the decision nodes
    assert "search.query.check_result" not in tree.decision_nodes

    # but the tool is still in the tree
    assert "send_email" in tree.tools

    # check the tree.tree is correct
    assert (
        "check_result"
        not in tree.tree["options"]["search"]["options"]["query"]["options"]
    )
